#!/usr/bin/env python3
"""
Component mixin exceptions.
"""

from ..component.exceptions import ComponentError


REQUIRE_KWARGS_SIGNATURE = ["export", "output_dir", "exist_ok"]


class InvalidProcessSignatureError(ComponentError):
    """
    Invalid process signature.

    Raise when the required kwargs are not met in the process method.
    """


class InvalidProvideSignatureError(ComponentError):
    """
    Invalid provide signature.
    Raise when the required kwargs are not met in the provide method.
    """


class InvalidConsumeSignatureError(ComponentError):
    """
    Invalid consume signature.
    Raise when the required kwargs are not met in the consume method.
    """



#!/usr/bin/env python3
"""Optional Dagster imports."""

from .exceptions import RunnerDependencyError

try:
    from dagster import (
        asset,
        AssetIn,
        Definitions,
        define_asset_job,
        mem_io_manager,
    )
except ImportError as err:
    raise RunnerDependencyError("dagster") from err


__all__ = [
    "asset",
    "AssetIn",
    "Definitions",
    "define_asset_job",
    "mem_io_manager",
]



#!/usr/bin/env python3
"""Optional Hydra imports."""

from .exceptions import RunnerDependencyError

try:
    from hydra import main as hydra_main
except ImportError as err:
    raise RunnerDependencyError("omegaconf") from err


__all__ = ["hydra_main"]



#!/usr/bin/env python3
"""
Dynamic imports.
Largely inspired from a code made by Mike Rye.
"""

import importlib
import inspect
import logging
import pathlib
import sys
from collections import namedtuple

LOGGER = logging.getLogger(__name__)


ModuleInfo = namedtuple("ModuleInfo", ["name", "ispkg"])


def find_modules(search_path, base_package):
    """
    Find Python modules recursively and generate valid module names.

    Args:
        search_path (pathlib.Path): Path to search for modules
        base_package (str): Base package name for imports

    Returns:
        List[str]: List of fully qualified module names

    Raises:
        ImportError: If the search path doesn't exist or isn't a valid Python package
    """
    search_path = pathlib.Path(search_path).resolve()
    if not search_path.exists():
        raise ImportError(f"Module path not found: {search_path}")

    # Verify this is a valid Python package
    if not (search_path / "__init__.py").exists():
        raise ImportError(
            f"Not a valid Python package (missing __init__.py): {search_path}"
        )

    # Find the package root by looking for the first parent with __init__.py
    package_root = search_path
    while (
        package_root.parent.exists() and (package_root.parent / "__init__.py").exists()
    ):
        package_root = package_root.parent

    modules = []
    for file in search_path.rglob("*.py"):
        if file.name == "__init__.py":
            continue

        try:
            # Get the module path relative to the package root
            relative_path = file.relative_to(package_root)
            # Convert path parts to module name (e.g., path/to/module.py -> path.to.module)
            module_parts = list(relative_path.with_suffix("").parts)

            # Construct the full module name
            if base_package:
                module_parts.insert(0, base_package)

            module_name = ".".join(module_parts)
            modules.append(module_name)

        except ValueError as e:
            LOGGER.warning("Failed to compute module name for %s: %s", file, e)
            continue

    return modules


def import_types(suffix, base_class, submod="type"):
    """
    Import types from submodules dynamically.

    Args:
        suffix (str): Class name suffix to match (e.g., "Generator")
        base_class (type): Base class whose location will be used to find types
        submod (str): Subdirectory name to search in (default: "type")

    Yields:
        type: Classes that match the specified suffix and inherit from base_class

    Raises:
        ImportError: If the submodule path is not found or is not a valid Python package
    """
    base_path = pathlib.Path(inspect.getfile(base_class)).parent
    base_module = base_class.__module__.split(".")[0]  # Get root package name

    submod_path = (base_path / submod).resolve()
    if not submod_path.exists():
        raise ImportError(f"Submodule path not found: {submod_path}")

    LOGGER.debug(
        "Importing types from %s (resolved to %s)",
        submod,
        submod_path,
    )

    for module_name in find_modules(submod_path, base_module):
        try:
            LOGGER.debug("Importing %s", module_name)
            mod = importlib.import_module(module_name)

            for name, cls in inspect.getmembers(mod):
                if (
                    name.endswith(suffix)
                    and cls != base_class
                    and inspect.isclass(cls)
                    and issubclass(cls, base_class)
                ):
                    LOGGER.debug("Found %s", name)
                    yield cls

        except ImportError as e:
            LOGGER.warning("Failed to import %s: %s", module_name, e)


def reload_self():
    """
    Reload the calling module and update all references.

    Returns:
        Reloaded class object
    """
    stack = inspect.stack()
    prev_frame = stack[1]
    cls = prev_frame[0].f_locals["cls"]
    mod_name = cls.__module__

    mod = sys.modules[mod_name]
    LOGGER.debug("Reloading module: %s", mod)
    importlib.reload(mod)
    reloaded_cls = getattr(mod, cls.__name__)

    for frame in stack:
        for namespace in ("f_locals", "f_globals"):
            try:
                ns_vars = getattr(frame[0], namespace)
            except AttributeError:
                continue

            for key, val in list(ns_vars.items()):
                if val is cls:
                    ns_vars[key] = reloaded_cls

    return reloaded_cls



#!/usr/bin/env python3
"""
Base class for all classes with cachable methods.
Cacheable inherits from jrai_common_mixins.cachable. 
In addition, it provides instance-level caching control.
"""

import dataclasses
import logging
import functools
import threading
from typing import Optional

from pydantic.dataclasses import dataclass
from pydantic.types import PositiveInt
from jrai_common_mixins.cachable import Cachable as BaseCachable

from ..dataclass.decorators.exportable.decorator import exportable

LOGGER = logging.getLogger(__name__)


@dataclass
class TimeOut:
    """
    Configuration for timeout.

    Attributes:
        seconds (int): Timeout in seconds.
        minutes (int): Timeout in minutes.
        hours (int): Timeout in hours.
        days (int): Timeout in days.
    """

    seconds: int = 0
    minutes: int = 0
    hours: int = 0
    days: int = 0

    def __bool__(self):
        """
        Check if the timeout is zero.
        """
        return not all(
            (self.seconds == 0, self.minutes == 0, self.hours == 0, self.days == 0)
        )


@exportable(stem_file="caching")
@dataclass
class CacheConfig:
    """
    Configuration for caching.

    Attributes:
        active (bool): Activates or deactivates caching.
        timeout (TimeOut): Timeout for cache items. If zero, no timeout applies.
        timeout_by_atime (bool): Whether to use atime or mtime for cache timeout.
        check_interval (PositiveInt): Interval to check for cache timeouts.
        cm_maxsize (Optional[int]): Max number of items in the cache to maintain.
    """

    active: bool = True
    timeout: TimeOut = dataclasses.field(default_factory=TimeOut)
    timeout_by_atime: bool = True
    check_interval: PositiveInt = 3600  # in seconds = 1 hour
    cm_maxsize: Optional[int] = None

    def __post_init__(self):
        if isinstance(self.timeout, dict):
            self.timeout = TimeOut(**self.timeout)  # pylint: disable=not-a-mapping

    def apply_to_environment(self):
        """
        Apply the configuration in my environment.
        """
        # pylint: disable=protected-access
        Cachable._update_global_cache_config(self)


class Cachable(BaseCachable):
    """
    A mixin that adds both global and instance-specific cache control.
    Inherits from jrai_common_mixins.cachable.
    """

    _GLOBAL_CACHE_CONFIG = CacheConfig()

    def __init__(self):
        """
        Initialize cache configuration for this instance.
        Instance-level settings are set to None by default,
        meaning the global settings will be used.
        """
        super().__init__()
        self._cache_config = None
        self._cleaner_thread = None
        self._cleaner_thread_stop_event = threading.Event()

        # Start cache cleaner if needed
        if self.is_cache_active() and self.cache_config.timeout:
            LOGGER.debug("Starting cache cleaner for instance.")
            self._start_cache_cleaner()

    def is_cache_active(self):
        """
        Returns True if caching is active.
        """
        return self.cache_config.active

    @property
    def cache_config(self):
        """
        Returns the cache configuration for this instance.
        """
        if self._cache_config is None:
            self._cache_config = self._GLOBAL_CACHE_CONFIG
        return self._cache_config

    @cache_config.setter
    def cache_config(self, cache_config):
        """
        Set the cache configuration for this instance.
        """
        if not isinstance(cache_config, CacheConfig):
            raise ValueError(f"Invalid cache config: {cache_config!r}")
        self._cache_config = cache_config
        self._manage_cache_cleaner_thread()

    def activate_cache(self):
        """
        Enable caching specifically for this instance.
        """
        self._cache_config.active = True
        self._manage_cache_cleaner_thread()

    def deactivate_cache(self):
        """
        Disable caching specifically for this instance.
        """
        self._cache_config.active = False
        self._stop_cache_cleaner()
        self.clear_cache()

    def _manage_cache_cleaner_thread(self):
        """
        Restarts the cleaner thread based on instance or global settings.
        """
        if self.is_cache_active() and self.cache_config.timeout:
            self._restart_cache_cleaner()
        else:
            self._stop_cache_cleaner()

    def _start_cache_cleaner(self):
        """
        Start a background thread to clear old cache entries.
        """
        if self._cleaner_thread and self._cleaner_thread.is_alive():
            return  # The thread is already running

        self._cleaner_thread_stop_event.clear()

        def cleaner():
            while not self._cleaner_thread_stop_event.is_set():
                threading.Event().wait(self.cache_config.check_interval)
                if self.cache_config.timeout:
                    LOGGER.debug("Timeout reached. Clearing old cache entries.")
                    self.clear_old_cache_entries(
                        by_atime=self.cache_config.timeout_by_atime,
                        **self.cache_config.timeout.__dict__,
                    )

        self._cleaner_thread = threading.Thread(target=cleaner, daemon=True)
        self._cleaner_thread.start()

    def _stop_cache_cleaner(self):
        """
        Stops the cache cleaner thread for this instance.
        """
        if self._cleaner_thread and self._cleaner_thread.is_alive():
            self._cleaner_thread_stop_event.set()
            self._cleaner_thread.join()

    def _restart_cache_cleaner(self):
        """
        Stop and restart the cache cleaner thread based on updated settings.
        """
        self._stop_cache_cleaner()
        self._start_cache_cleaner()

    @classmethod
    def caching_property(cls, method):
        """
        Decorator to cache the result of a property.
        Based on the caching_property implemented in jrai_common_mixins.cachable
        """

        @property
        @functools.wraps(method)
        def wrapper(self):
            if self.is_cache_active():
                return super().caching_property(method).fget(self)
            return method(self)

        return wrapper

    @classmethod
    def caching_method(cls, *arg_cmps, _cm_maxsize=None, **kwarg_cmps):
        """
        Decorator to cache the result of a method.
        Based on the caching_method implemented in jrai_common_mixins.cachable.
        """

        def decorator(method):
            @functools.wraps(method)
            def cmeth(self, *args, **kwargs):
                cache_activ = self.is_cache_active()
                if cache_activ:
                    cm_maxsize = (
                        _cm_maxsize
                        if _cm_maxsize is not None
                        else self.cache_config.cm_maxsize
                    )
                    return super().caching_method(
                        *arg_cmps, _cm_maxsize=cm_maxsize, **kwarg_cmps
                    )(method)(self, *args, **kwargs)
                return method(self, *args, **kwargs)

            return cmeth

        return decorator

    @classmethod
    def _update_global_cache_config(cls, config):
        """
        Update the global cache configuration.
        """
        cls._GLOBAL_CACHE_CONFIG = config



#!/usr/bin/env python3
"""
Configuration for custom processors.
"""

from pydantic.dataclasses import dataclass

from .....dataclass.decorators.registry.decorator import registry
from .....dataclass.decorators.exportable.decorator import exportable

from ..base.processor import CustomProcessor
from .base import CustomConfig


@registry(base_cls=CustomProcessor)
@exportable(strategy="registry")
@dataclass
class CustomProcessorConfig(CustomConfig):
    """Custom processor configuration."""



#!/usr/bin/env python3
"""
Base export strategy.
"""

import dataclasses
import logging

from pydantic.dataclasses import dataclass

from ..base import ExportStrategy
from ...decorator import exportable
from .....format.info import has_required_fields
from ......fallback.typing import NoneType


LOGGER = logging.getLogger(__name__)


class RegistryExportStrategy(ExportStrategy, register_name="registry"):
    """Export strategy for registry-based configurations."""

    @classmethod
    def export(  # pylint: disable=too-many-arguments
        cls,
        data_cls,
        output_dir,
        format_type="yaml",
        exist_ok=False,
        makedirs=False,
        **kwargs,
    ):
        """Export strategy for registry dataclasses."""
        # pylint: disable=protected-access
        registry_class = data_cls._REG_BASE_CLASS_
        cls_settings_attr = data_cls._SETUP_.settings_dataclass_attr

        data_cls.reload_registry()  # Reload registry

        # Export schema for each registered type
        for registration_name, registered_cls in registry_class._get_register().items():
            settings_type = getattr(registered_cls, cls_settings_attr, None)
            temp_config = cls._create_exportable_temp_config(
                data_cls, registration_name, settings_type
            )
            temp_config.export(
                output_dir,
                format_type=format_type,
                exist_ok=exist_ok,
                makedirs=makedirs,
                **kwargs,
            )

    @classmethod
    def _create_exportable_temp_config(cls, data_cls, registration_name, settings_type):
        """Create a temporary config class for export."""
        # pylint: disable=protected-access
        settings_attr = data_cls._SETUP_.settings_attr
        register_name_attr = data_cls._SETUP_.register_name_attr

        annotations = dict(data_cls.__annotations__)

        # Special case for NoneType
        if settings_type is None or settings_type is NoneType:
            LOGGER.debug(
                "Skipping settings for %s as it uses NoneType", registration_name
            )
            # Use None as a placeholder for settings_type
            annotations[settings_attr] = type(None)
        else:
            annotations[settings_attr] = settings_type

        annotations[register_name_attr] = str

        @exportable
        @dataclass
        class TempConfig(data_cls):  # pylint: disable=C0115
            _EXPORT_SETUP_ = data_cls._EXPORT_SETUP_
            __annotations__ = annotations

            # pylint: disable=E3701
            locals()[register_name_attr] = dataclasses.field(
                default=registration_name,
                metadata={"_FORCE_REQUIRED_": True},
            )

            # Only add settings field if not NoneType
            if settings_type is not None and settings_type is not NoneType:
                if has_required_fields(settings_type):
                    locals()[settings_attr] = dataclasses.field(
                        default=None,
                        metadata={"_FORCE_REQUIRED_": True},
                    )

        # Update class setup and metadata
        TempConfig._EXPORT_SETUP_.stem_file = registration_name
        TempConfig.__name__ = cls.__name__
        TempConfig.__qualname__ = cls.__qualname__
        TempConfig.__module__ = cls.__module__
        TempConfig.__doc__ = data_cls.__doc__

        return TempConfig



#!/usr/bin/env python3
"""
Base formatter.
"""

from abc import ABC, abstractmethod

from ...mixin.registrable import Registrable
from ...mixin.settings import SettingsMixin


class Formatter(ABC, Registrable, SettingsMixin):
    """Base formatter."""

    _REGISTER = {}
    _TYPE_SUBMODULE = "type"

    def __init__(self, settings):
        Registrable.__init__(self)
        SettingsMixin.__init__(self, settings)

    @abstractmethod
    def to_str(self, class_info):
        """Format the class information."""

    @abstractmethod
    def display(self, content):
        """Display the formatted content."""

    @abstractmethod
    def to_serialized(self, class_info):
        """Convert DataclassInfo to a serialized object."""



#!/usr/bin/env python3
"""
Setup for registry dataclass decorator.
"""

from pydantic.dataclasses import dataclass


@dataclass
class RegistrySetup:
    """Registry Setup."""

    register_name_attr: str = "name"
    settings_attr: str = "settings"
    settings_dataclass_attr: str = "SETTINGS_DATACLASS"



#!/usr/bin/env python3
"""Fallback for typing.Protocol."""


class DummyProtocol:
    """
    A dummy implementation to replace typing.Protocol for Python 3.7 compatibility.
    """

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)


def runtime_checkable(cls):
    """
    A dummy decorator to mimic typing.runtime_checkable.
    """
    return cls


# Fallback mechanism
try:
    from typing import Protocol, runtime_checkable  # pylint: disable=unused-import
except ImportError:
    Protocol = DummyProtocol
    # runtime_checkable already defined above



#!/usr/bin/env python3
"""
Kwargs utils.
"""

import inspect


def validate_kwargs_signature(
    obj, method_name, required_params, exception_type=TypeError
):
    """
    Validate that the method of a given object has the required parameters in its signature.

    Args:
        obj: The object containing the method to validate.
        method_name (str): Name of the method to inspect.
        required_params (list): List of required keyword argument names.
        exception_type (Exception): Type of exception to raise if validation fails.

    Raises:
        exception_type: If the method's signature does not contain the required parameters.
    """
    method = getattr(obj, method_name, None)
    if method is None:
        raise AttributeError(
            f"{obj.__class__.__name__} has no method named '{method_name}'"
        )

    signature = inspect.signature(method)
    func_params = signature.parameters

    missing_params = [param for param in required_params if param not in func_params]
    if missing_params:
        raise exception_type(
            f"Method '{method_name}' in class '{obj.__class__.__name__}' "
            f"is missing required parameters: {missing_params}"
        )



#!/usr/bin/env python3
"""Runner-specific optional dependency errors."""

from ..exceptions import BaseOptionalDependencyError


class RunnerDependencyError(BaseOptionalDependencyError):
    """Error raised when a Runner-related optional dependency is missing."""

    def __init__(self, package_name: str):
        super().__init__(package_name, extra_name="runner")



#!/usr/bin/env python3
"""Unit tests for the @wenv decorator."""

import unittest
from typing import Dict, Any
from pydantic.dataclasses import dataclass

from pypassist.dataclass.decorators.wenv import wenv, WenvSetupError
from pypassist.dataclass.decorators.wenv.setup import WenvSetup


class TestWenvDecorator(unittest.TestCase):
    """Test cases for the @wenv decorator functionality."""

    def test_basic_decoration(self):
        """Test basic decoration of a dataclass."""

        @wenv
        @dataclass
        class NestedConfig:
            """Nested configuration class."""

            value: str

        @wenv
        @dataclass
        class TestConfig:
            """Test configuration class."""

            name: str
            settings: Dict[str, Any]
            nested: Dict[str, NestedConfig]

        # Create an instance with nested dictionary
        config = TestConfig(
            name="test",
            settings={"param": "value"},
            nested={
                "item1": NestedConfig(value="test1"),
                "item2": {"value": "test2"},  # This should be converted to NestedConfig
            },
        )

        # Verify the class is properly decorated
        self.assertTrue(hasattr(TestConfig, "_SETUP_"))
        self.assertTrue(hasattr(TestConfig, "_WENV_"))

        # Verify dictionary field processing
        self.assertIsInstance(config.nested["item1"], NestedConfig)
        self.assertIsInstance(config.nested["item2"], NestedConfig)
        self.assertEqual(config.nested["item2"].value, "test2")

    def test_invalid_setup(self):
        """Test handling of invalid setup configuration."""
        with self.assertRaises(WenvSetupError):

            @wenv(setup=123)  # Invalid setup type - should be WenvSetup instance
            @dataclass
            class TestConfig:
                """Test configuration class."""

                settings: Dict[str, Any]


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Working environment exceptions.
"""

from ...exceptions import DataclassError


class WenvError(DataclassError):
    """Base exception for work environment related errors."""


class WenvSetupError(WenvError):
    """Exception for work environment setup errors."""



#!/usr/bin/env python3
"""
Custom work environment.
"""

from ...workenv.base.workenv import BaseWorkEnv
from ....mixin.cachable import Cachable


class CustomWorkEnv(BaseWorkEnv, Cachable):
    """Custom work environment."""

    def __init__(self, config, parent_env=None):
        BaseWorkEnv.__init__(self, config, parent_env)
        Cachable.__init__(self)

    @Cachable.caching_property
    def sinks(self):
        """
        The dict of instantiated custom sinks.
        """
        return self._get_custom_components("sinks")

    @Cachable.caching_property
    def sources(self):
        """
        The dict of instantiated custom sources.
        """
        return self._get_custom_components("sources")

    @Cachable.caching_property
    def processors(self):
        """
        The dict of instantiated custom processors.
        """
        return self._get_custom_components("processors")

    def _get_custom_components(self, section_name):
        """
        Get custom components from nested sections.
        """
        section = getattr(self.config, section_name)
        return {
            name: conf.get_component(workenv=self.parent_env)
            for (name, conf) in section.items()
        }



#!/usr/bin/env python3
"""
Test cachable.py.
"""


import unittest
import time
import random
import logging
import datetime
from pypassist.mixin.cachable import Cachable, CacheConfig

# Comments:
# freezegun does not work with deamon threads
# that's why we use time.sleep instead of freezegun


class MyCachable(Cachable):
    """
    Test subclass for Cachable to use in unit tests.
    """

    @Cachable.caching_property
    def cached_property(self):
        """
        Dummy caching property.
        Returns the current UTC time.
        """
        return datetime.datetime.now(datetime.timezone.utc)

    @Cachable.caching_method()
    def cached_method(self, x, y=0):
        """
        Dummy caching method.
        Returns the sum of x and y and a random number.
        """
        z = random.random()
        return x + y + z


class TestCachable(unittest.TestCase):
    """
    Test Cachable class.
    """

    def setUp(self):
        """Initialize for each test."""
        self.instance = MyCachable()
        ## reset the cache settings
        self.instance.cache_config = CacheConfig()

    def test_cached_property(self):
        """Test the caching of the cached_property."""
        value1 = self.instance.cached_property
        value2 = self.instance.cached_property
        self.assertEqual(value1, value2)  # The property should return the cached value

    def test_cached_method(self):
        """Test the caching of the cached_method."""
        result1 = self.instance.cached_method(3, 4)
        result2 = self.instance.cached_method(3, 4)  # Should hit the cache
        self.assertEqual(result1, result2)  # Should return the same result from cache

    # ------------------- Test the cache settings ------------------------------

    def test_cache_enabled_for_property(self):
        """Test enabling and disabling the cache for the cached_property."""
        # Disable caching
        self.instance.deactivate_cache()
        result1 = self.instance.cached_property  # This should not be cached
        result2 = self.instance.cached_property  # This should compute again
        self.assertNotEqual(result1, result2)  # Results should differ

        # Enable caching again
        self.instance.activate_cache()
        result3 = self.instance.cached_property  # This should be cached
        result4 = self.instance.cached_property  # This should hit the cache
        self.assertEqual(result3, result4)  # Results should be the same

    def test_cache_enabled_for_method(self):
        """Test enabling and disabling the cache for the cached_property."""
        # Disable caching
        self.instance.deactivate_cache()
        result1 = self.instance.cached_method(10, 10)  # This should not be cached
        result2 = self.instance.cached_method(10, 10)  # This should compute again
        self.assertNotEqual(result1, result2)  # Results should differ

        # Enable caching again
        self.instance.activate_cache()
        result3 = self.instance.cached_method(10, 12)  # This should be cached
        result4 = self.instance.cached_method(10, 12)  # This should hit the cache
        self.assertEqual(result3, result4)  # Results should be the same

    def test_cache_timeout_for_property(self):
        """Test setting the check interval for the cache cleaner."""
        # pylint: disable=protected-access
        self.instance.cache_config = CacheConfig(
            check_interval=1, timeout={"seconds": 1}
        )
        result1 = self.instance.cached_property  # Cache this result
        result2 = self.instance.cached_property  # Should hit the cache
        self.assertEqual(result1, result2)  # Results should be the same
        time.sleep(3)  # Wait for the cache to expire
        result3 = self.instance.cached_property
        self.assertNotEqual(result1, result3)  # Results should differ

    def test_cache_timeout_for_method(self):
        """Test setting a cache timeout."""
        # pylint: disable=protected-access
        self.instance.cache_config = CacheConfig(
            check_interval=1, timeout={"seconds": 1}
        )
        result1 = self.instance.cached_method(10, 10)  # This should cache the result
        result2 = self.instance.cached_method(10, 10)  # This should hit the cache
        self.assertEqual(result1, result2)  # Results should be the same
        time.sleep(3)  # Wait for the cache to expire
        result3 = self.instance.cached_method(10, 10)  # This should compute again
        self.assertNotEqual(result1, result3)

    # ------------------- Test GLOBAL cache settings ------------------------------

    def test_global_cache_config_update(self):
        """Test setting a cache timeout."""
        CacheConfig(active=False).apply_to_environment()
        tc = MyCachable()
        self.assertFalse(tc.is_cache_active())


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    unittest.main()



#!/usr/bin/env python3
"""Optional OmegaConf imports."""

from .exceptions import RunnerDependencyError

try:
    from omegaconf import OmegaConf
except ImportError as err:
    raise RunnerDependencyError("omegaconf") from err

__all__ = ["OmegaConf"]



#!/usr/bin/env python3
# pylint: disable=import-self
"""
Typing functions for retrocompatibility with older Python versions.
"""
import logging
import sys
from typing import Generic, _GenericAlias

LOGGER = logging.getLogger(__name__)

if sys.version_info >= (3, 9):
    Dict = dict
    List = list
else:
    LOGGER.debug("Using and old definition of Dict")
    from typing import Dict, List


def _log_definition(name, err):
    """
    Load the definition of a function.

    Args:
        name:
            The name of the function begin defined.

        err:
            The import error.
    """
    LOGGER.debug(
        'Defining a simplified version of the function "%s" because'
        "the typing package provided by Python %s does not define it: %s",
        name,
        sys.version,
        err,
    )


try:
    from typing import get_origin
except ImportError as err:
    _log_definition("get_origin", err)

    # https://github.com/python/cpython/blob/6586b171ea842151c24d2228d06a69d2fecaf29f/Lib/typing.py
    def get_origin(tp):
        """
        Simplified version of typing.get_origin() as defined in Python 3.12 for
        older version of Python.
        """
        if tp is Generic:
            return Generic
        return getattr(tp, "__origin__", None)


try:
    from typing import get_args
except ImportError as err:
    _log_definition("get_args", err)

    import collections.abc

    # https://github.com/python/cpython/blob/6586b171ea842151c24d2228d06a69d2fecaf29f/Lib/typing.py

    def _is_param_expr(arg):
        """
        Simplified implementation of typing._is_param_expr from Python 3.12.
        """
        return arg is ... or isinstance(arg, (tuple, list))

    def _should_unflatten_callable_args(typ, args):
        """
        Implementation of typing._should_unflatten_callable_args from Python 3.12.
        """
        return typ.__origin__ is collections.abc.Callable and not (
            len(args) == 2 and _is_param_expr(args[0])
        )

    def get_args(tp):
        """
        Simplified version of typing.get_args() as defined in Python 3.12 for
        older version of Python.
        """
        try:
            return (tp.__origin__,) + tp.__metadata__
        except AttributeError:
            pass
        if isinstance(tp, _GenericAlias):
            res = tp.__args__
            if _should_unflatten_callable_args(tp, res):
                res = (list(res[:-1]), res[-1])
            return res

        return getattr(tp, "__args__", ())


def type_matches_annotation(value, annotation):
    """
    Check if the type of a value matches an annotation. This function will catch
    TypeErrors raised by older versions of Python and use a workaround.

    Args:
        value:
            The value to check.

        annotation:
            The matching annotation.

    Returns:
        True if the type matches the annotation, else False.
    """
    try:
        return isinstance(value, annotation)
    except TypeError:
        args = get_args(annotation)
        return isinstance(value, args)


NoneType = type(None)


__all__ = [
    "Dict",
    "List",
    "get_args",
    "get_origin",
    "type_matches_annotation",
    "NoneType",
]



#!/usr/bin/env python3
"""
Base work environment functionality.
"""

from ....dataclass.decorators.wenv.validation import is_wenv

from ..component.exceptions import InvalidComponentError


class BaseWorkEnv:
    """Base class for work environments."""

    def __init__(self, config, parent_env=None):
        self.config = config
        self._validate_config()
        self.parent_env = parent_env

    def _validate_config(self):
        if not is_wenv(self.config):
            raise ValueError(
                "Invalid work environment configuration. "
                "Expected a pydantic dataclass decorated with @wenv."
            )

    def get_component(self, path):
        """
        Get component by path.

        Args:
            path: Component path in format "section.name"

        Returns:
            Component instance

        Raises:
            ValueError: If component path is invalid or component not found
        """
        try:
            section_name, component_name = path.split(".")
        except ValueError as err:
            raise ValueError(
                f"Invalid component path: {path}. Expected format: 'section.name'"
            ) from err

        section = getattr(self, section_name)
        if component_name not in section:
            raise InvalidComponentError(
                f"{self.__class__.__name__}: Component '{component_name}' not found in section '{section_name}'. "
                f"Available components: {', '.join(section.keys())}"
            )

        return section[component_name]



#!/usr/bin/env python3
"""
Sink component protocol for workflow module.
"""

from .base import WorkflowComponent
from ....fallback.protocol import runtime_checkable, Protocol


@runtime_checkable
class SinkComponent(WorkflowComponent, Protocol):
    """Protocol for components that consume data."""

    def consume(self, **kwargs):
        """Consume  data with flexible signature.

        All implementations should support these keyword arguments:
            export (bool): Whether to export the results
            output_dir (str): Directory to export results to
            exist_ok (bool): Whether to overwrite existing files

        Args:
            **kwargs: Arbitrary keyword arguments

        Returns:
            None
        """



#!/usr/bin/env python3

"""
Export utilities.
"""

import logging
import pathlib

from ..exceptions import PypassistError

LOGGER = logging.getLogger(__name__)


## -- Exceptions
class ExportError(PypassistError):
    """Base exception for export utilities."""


class DirectoryNotFoundError(ExportError):
    """Raised when the parent directory for export doesn't exist."""


## -- Utilities


def check_parent_dir_exists(filepath):
    """
    Check if the parent directory exists without creating it.

    Args:
        filepath: Path to check
    Returns:
        bool: True if parent directory exists, False otherwise
    """
    filepath = pathlib.Path(filepath)
    return filepath.parent.exists()


def get_valid_filepath(filepath, exist_ok, makedirs):
    """
    Validate and prepare a file path before use.

    This function ensures that the given file path is valid by resolving it to an absolute path.
    It also verifies that the parent directory exists and optionally creates it if `makedirs` is True.
    If the file already exists and `exist_ok` is False, an exception is raised.

    Args:
        filepath (str or pathlib.Path): The path to validate and prepare.
        exist_ok (bool, optional): If False, raises an error if the file already exists. Defaults to False.
        makedirs (bool, optional): If True, creates the parent directory if it doesn't exist. Defaults to False.

    Raises:
        DirectoryNotFoundError: If the parent directory does not exist and `makedirs` is False.
        FileExistsError: If the file already exists and `exist_ok` is False.

    Returns:
        pathlib.Path: The resolved and validated file path.
    """
    filepath = pathlib.Path(filepath).resolve()

    if makedirs:
        create_directory(filepath.parent)

    if not check_parent_dir_exists(filepath):
        raise DirectoryNotFoundError(
            f"Parent directory does not exist: {filepath.parent}"
        )

    if filepath.exists() and not exist_ok:
        raise FileExistsError(
            f"File already exists: {filepath} (use exist_ok=True to overwrite)."
        )
    return filepath


def export_to_csv(
    dataframe, filepath, sep="\t", exist_ok=False, makedirs=False, **kwargs
):
    """
    Export DataFrame to CSV, assuming parent directory exists.

    Args:
        dataframe: DataFrame to export
        filepath: Output filepath
        sep: Separator to use, defaults to tab
        **kwargs: Additional arguments for pandas.to_csv()

    Raises:
        DirectoryNotFoundError: If parent directory doesn't exist
        ExportError: If there's an error during export
    """
    filepath = get_valid_filepath(filepath, exist_ok, makedirs)

    LOGGER.debug("Exporting DataFrame to: %s", filepath)
    dataframe.to_csv(filepath, sep=sep, **kwargs)


def export_string(string, filepath, exist_ok=False, encoding="utf-8", makedirs=False):
    """
    Export a string to a file.

    Args:
        string (str): The string to export
        filepath (str): Output filepath
        exist_ok (bool): Whether to overwrite existing file
        encoding (str): Encoding to use
        makedirs (bool): Whether to create parent directories

    Raises:
        DirectoryNotFoundError: If parent directory doesn't exist
        ExportError: If there's an error during export
    """
    filepath = get_valid_filepath(filepath, exist_ok, makedirs)

    with open(filepath, "w", encoding=encoding) as file:
        file.write(string)

    LOGGER.debug("Exporting string to file: %s", filepath)


def create_directory(directory_path, mode=0o755, exist_ok=True, parents=True):
    """
    Create a directory.

    Args:
        directory_path (str): Path to directory
        mode (int): Permissions to set on the directory
        exist_ok (bool): Whether existing directory is acceptable
        parents (bool): Whether to create parent directories if they don't exist

    Returns:
        Path: The path to the created directory
    """
    path = pathlib.Path(directory_path).resolve()
    path.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
    return path



#!/usr/bin/env python3
"""Unit tests for custom workflow operators."""

import tempfile
import unittest
from pathlib import Path

from pydantic.dataclasses import dataclass

from pypassist.runner.workenv.custom.base.processor import CustomProcessor
from pypassist.runner.workenv.custom.base.source import CustomSource
from pypassist.dataclass.decorators.viewer import viewer
from pypassist.runner.workenv.mixin.exceptions import InvalidProcessSignatureError
from pypassist.utils.kwargs import validate_kwargs_signature


class TestCustomOperator(unittest.TestCase):
    """Test cases for custom workflow operators."""

    def setUp(self):
        """Set up test fixtures."""
        self.temp_dir = tempfile.mkdtemp()
        self.output_dir = Path(self.temp_dir)

    def test_basic_operator(self):
        """Test basic custom operator functionality."""

        @viewer
        @dataclass
        class ReverserSettings:
            """Settings for string reversal."""

            capitalize: bool = False

        class Reverser(CustomProcessor, register_name="reverse"):
            """A simple string reverser operator."""

            SETTINGS_DATACLASS = ReverserSettings

            def __init__(self, settings=None):
                if settings is None:
                    settings = ReverserSettings()
                CustomProcessor.__init__(self, settings)

            @classmethod
            def init_from_config(cls, config, workenv=None):
                """Initialize the component from configuration."""
                return cls(config.settings)

            def __call__(self, text):
                """Reverse the input text."""
                result = text[::-1]
                if self.settings.capitalize:
                    result = result.upper()
                return result

            def process(
                self, text=None, export=False, output_dir=None, exist_ok=True, **kwargs
            ):
                """Process the text and return the processed text.

                Args:
                    text: The text to process
                    export: Whether to export the results
                    output_dir: Directory to export results to
                    exist_ok: Whether to overwrite existing files
                    **kwargs: Additional keyword arguments

                Returns:
                    The processed text
                """
                if text is None:
                    raise ValueError("Text parameter is required")

                result = self(text)
                if export and output_dir:
                    # Export result
                    result_file = output_dir / f"{self.__class__.__name__.lower()}.txt"
                    result_file.parent.mkdir(parents=True, exist_ok=exist_ok)
                    result_file.write_text(result)

                    # Export settings
                    settings_str = self.settings.to_str(format_type="yaml")
                    settings_file = (
                        output_dir / f"{self.__class__.__name__.lower()}_settings.txt"
                    )
                    settings_file.write_text(settings_str)
                return result

        # Test basic operation
        reverser = Reverser(ReverserSettings(capitalize=False))
        self.assertEqual(reverser("hello"), "olleh")

        # Test with settings
        reverser = Reverser(ReverserSettings(capitalize=True))
        self.assertEqual(reverser("hello"), "OLLEH")

        # Test asset generation
        asset_func = reverser.get_assetable_func()
        result = asset_func(text="hello", export=True, output_dir=self.output_dir)
        self.assertEqual(result, "OLLEH")

        # Verify exported files
        result_file = self.output_dir / "reverser.txt"
        settings_file = self.output_dir / "reverser_settings.txt"
        self.assertTrue(result_file.exists())
        self.assertTrue(settings_file.exists())
        self.assertEqual(result_file.read_text(), "OLLEH")

    def test_operator_registration(self):
        """Test operator registration and retrieval."""

        @viewer
        @dataclass
        class DummySettings:
            """Dummy settings."""

            value: str = "test"

        class DummyOperator(CustomSource, register_name="dummy"):
            """A dummy operator for testing registration."""

            SETTINGS_DATACLASS = DummySettings

            def __init__(self, settings=None):
                if settings is None:
                    settings = DummySettings()
                CustomSource.__init__(self, settings)

            @classmethod
            def init_from_config(cls, config, workenv=None):
                """Initialize the component from configuration."""
                return cls(config.settings)

            def __call__(self, text):
                return f"{text}_{self.settings.value}"

            def provide(self, **kwargs):
                """Provide data with flexible signature.

                Args:
                    **kwargs: Arbitrary keyword arguments including:
                        - text: The text to provide
                        - export: Whether to export the results
                        - output_dir: Directory to export results to
                        - exist_ok: Whether to overwrite existing files

                Returns:
                    The provided data
                """
                text = kwargs.get("text")
                if text is None:
                    raise ValueError("Text parameter is required")
                return self(text)

        # Verify registration
        self.assertTrue(hasattr(DummyOperator, "_REGISTER"))
        self.assertIn("dummy", DummyOperator._REGISTER)

        # Test operator retrieval
        retrieved = CustomSource.get_registered("dummy")
        self.assertEqual(retrieved, DummyOperator)

        # Test operator functionality
        operator = DummyOperator()
        self.assertEqual(operator("input"), "input_test")

    def test_invalid_operator(self):
        """Test handling of invalid operator configurations."""
        with self.assertRaises(TypeError):
            # Missing required methods
            class InvalidOperator(CustomSource, register_name="invalid"):
                """An invalid operator missing required methods."""

                pass

            # This should raise TypeError due to missing required methods
            # pylint: disable=abstract-class-instantiated
            # pylint: disable=no-value-for-parameter
            InvalidOperator()

    def test_process_signature_validation(self):
        """Test process signature validation."""

        @viewer
        @dataclass
        class SignatureSettings:
            """Settings for signature testing."""

            value: str = "test"

        # Test with valid signature
        class ValidProcessor(CustomProcessor, register_name="valid_processor"):
            """A processor with a valid process signature."""

            SETTINGS_DATACLASS = SignatureSettings

            def __init__(self, settings=None):
                if settings is None:
                    settings = SignatureSettings()
                CustomProcessor.__init__(self, settings)

            @classmethod
            def init_from_config(cls, config, workenv=None):
                """Initialize the component from configuration."""
                return cls(config.settings)

            def __call__(self, text):
                return f"{text}_{self.settings.value}"

            def process(self, text, export=False, output_dir=None, exist_ok=True):
                """Process with valid signature including all required parameters.

                Args:
                    text: The text to process
                    export: Whether to export the results
                    output_dir: Directory to export results to
                    exist_ok: Whether to overwrite existing files

                Returns:
                    The processed text
                """
                result = self(text)
                if export and output_dir:
                    result_file = output_dir / f"{self.__class__.__name__.lower()}.txt"
                    result_file.parent.mkdir(parents=True, exist_ok=exist_ok)
                    result_file.write_text(result)
                return result

        # This should work fine
        processor = ValidProcessor()
        asset_func = processor.get_assetable_func()
        result = asset_func(
            text="hello", export=True, output_dir=self.output_dir, exist_ok=True
        )
        self.assertEqual(result, "hello_test")

        # Test with invalid signature
        class InvalidProcessor:
            """A class with an invalid process signature."""

            def __init__(self):
                pass

            def process(self, text):
                """Process method missing required parameters."""
                return text

        # This should raise InvalidProcessSignatureError
        invalid_processor = InvalidProcessor()
        with self.assertRaises(InvalidProcessSignatureError):
            validate_kwargs_signature(
                obj=invalid_processor,
                method_name="process",
                required_params=["export", "output_dir", "exist_ok"],
                exception_type=InvalidProcessSignatureError,
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Base exceptions for optional dependencies."""

import inspect
import importlib.metadata


class BaseOptionalDependencyError(ImportError):
    """Base class for optional dependency errors."""

    @staticmethod
    def _find_caller_package():
        """Determine the package that triggered the import."""
        for frame in inspect.stack():
            module = inspect.getmodule(frame[0])
            if module and not module.__name__.startswith("pypassist"):
                try:
                    package_name = module.__name__.split(".")[0]
                    importlib.metadata.metadata(package_name)
                    return package_name
                except importlib.metadata.PackageNotFoundError:
                    continue
        return "pypassist"

    def __init__(self, package_name: str, extra_name: str):
        self.missing_package = package_name
        self.extra_name = extra_name
        parent_package = self._find_caller_package()

        message = (
            f"{package_name} is not installed. Install {parent_package}[{extra_name}] "
            f"to use {package_name}-based features: pip install {parent_package}[{extra_name}]"
        )
        super().__init__(message)



#!/usr/bin/env python3
"""
Viewer decorator.
"""


import logging

from ..base import BaseDecorator
from .setup import ViewerSetup
from .exceptions import ViewerSetupError
from ...format.base import Formatter
from ...format.info import DataclassInfo

LOGGER = logging.getLogger(__name__)


def viewer(_cls=None, *, setup=None, **kwargs):
    """Viewer dataclass decorator for enhanced data visualization and formatting.

    Args:
        _cls: The class to decorate
        setup: ViewerSetup instance containing display configuration
        kwargs: Setup fields. Will override corresponding field in setup.

    Returns:
        Decorated class with visualization and formatting capabilities

    Raises:
        ViewerSetupError: If setup configuration is invalid

    Example:
        @viewer
        class MyConfig:
            name: str
            _private_field: str  # Will be hidden in output
    """

    def wrap(cls):
        try:
            cls, setup_instance = BaseDecorator.decorate_class(
                cls, ViewerSetup, setup, **kwargs
            )
        except ValueError as err:
            raise ViewerSetupError(str(err)) from err

        class ViewerClass(cls):  # pylint: disable=R0903, C0115
            cls._SETUP_ = setup_instance

            def to_info(self):
                """Convert the instance to DataclassInfo."""
                setup = self._SETUP_
                hide_private = getattr(setup, "hide_private", False)
                return DataclassInfo.from_instance(self, hide_private=hide_private)

            def to_str(self, format_type="yaml", **kwargs):
                """Format the class data to a string representation.

                Args:
                    format_type: Output format (yaml, json, etc.)
                    **kwargs: Additional formatting options

                Returns:
                    Formatted string representation
                """
                formatter = Formatter.get_registered(format_type)()
                formatter.update_settings(**kwargs)
                return formatter.to_str(self.to_info())

            def serialize(self, format_type="json", **kwargs):
                """Convert the class to a serialized object.

                Args:
                    format_type: Serialization format
                    **kwargs: Additional serialization options

                Returns:
                    Serialized representation
                """
                formatter = Formatter.get_registered(format_type)()
                formatter.update_settings(**kwargs)
                return formatter.to_serialized(self.to_info())

            def view(self, format_type="yaml", **kwargs):
                """Display formatted class information.

                Args:
                    format_type: Display format
                    **kwargs: Additional display options
                """
                content = self.to_str(format_type, **kwargs)
                formatter = Formatter.get_registered(format_type)()
                formatter.display(content)

        BaseDecorator.preserve_cls_metadata(ViewerClass, cls)
        return ViewerClass

    if _cls is None:
        return wrap
    return wrap(_cls)



#!/usr/bin/env python3
"""Package stub."""

from .viewer.decorator import viewer
from .viewer.setup import ViewerSetup
from .viewer.validation import is_viewer

from .wenv.decorator import wenv
from .wenv.validation import is_wenv

from .registry.decorator import registry
from .registry.setup import RegistrySetup

# from .registry.validation import is_registry

# from .exportable_registry.decorator import exportable_registry
# from .exportable_registry.setup import ExportableRegistrySetup

# from .exportable_registry.decorator import exportable_registry
# from .exportable_registry.setup import ExportableRegistrySetup

# from .exportable.decorator import exportable
# from .exportable.setup import ExportableSetup

# from .exportable.exportable.decorator import exportable
# from .exportable.setup import ExportableSetup

# from .exportable.exportable_registry.decorator import exportable_registry
# from .exportable.exportable_registry.setup import ExportableRegistrySetup

__all__ = [
    "viewer",
    "ViewerSetup",
    "is_viewer",
    "registry",
    "RegistrySetup",
    "wenv",
    "is_wenv",
]



#!/usr/bin/env python3
"""Unit tests for the @viewer decorator."""

import unittest
from typing import Optional

from pydantic.dataclasses import dataclass

from pypassist.dataclass.decorators.viewer import viewer
from pypassist.dataclass.decorators.viewer.setup import ViewerSetup


class TestViewerDecorator(unittest.TestCase):
    """Test cases for the @viewer decorator functionality."""

    def test_basic_decoration(self):
        """Test basic decoration of a dataclass."""

        @viewer
        @dataclass
        class TestConfig:
            """Test configuration class."""

            name: str
            value: int
            description: Optional[str] = None

        # Create an instance
        config = TestConfig(name="test", value=42)

        # Verify instance has viewer methods
        self.assertTrue(hasattr(config, "to_info"))
        self.assertTrue(hasattr(config, "to_str"))
        self.assertTrue(hasattr(config, "serialize"))
        self.assertTrue(hasattr(config, "view"))

    def test_string_conversion(self):
        """Test conversion to string formats."""

        @viewer
        @dataclass
        class TestConfig:
            """Test configuration class."""

            name: str
            value: int

        config = TestConfig(name="test", value=42)

        # Test YAML format
        yaml_str = config.to_str(format_type="yaml")
        self.assertIsInstance(yaml_str, str)
        self.assertIn("name: test", yaml_str)
        self.assertIn("value: 42", yaml_str)

        # Test JSON format
        json_str = config.to_str(format_type="json")
        self.assertIsInstance(json_str, str)
        self.assertIn('"name": "test"', json_str)
        self.assertIn('"value": 42', json_str)

    def test_serialization(self):
        """Test serialization functionality."""

        @viewer
        @dataclass
        class TestConfig:
            """Test configuration class."""

            name: str
            value: int

        config = TestConfig(name="test", value=42)

        # Test JSON serialization
        json_obj = config.serialize(format_type="json")
        self.assertIsInstance(json_obj, dict)
        self.assertEqual(json_obj["name"], "test")
        self.assertEqual(json_obj["value"], 42)

    def test_custom_setup(self):
        """Test custom setup configuration."""
        setup = ViewerSetup(hide_private=True)

        @viewer(setup=setup)
        @dataclass
        class TestConfig:
            """Test configuration class."""

            public_field: str
            _private_field: str

        config = TestConfig(public_field="public", _private_field="private")

        # Verify private field is hidden in string representation
        yaml_str = config.to_str(format_type="yaml")
        self.assertIn("public_field:", yaml_str)
        self.assertNotIn("_private_field:", yaml_str)

    def test_invalid_setup(self):
        """Test handling of invalid setup configuration."""
        with self.assertRaises(Exception):  # Should raise specific exception

            @viewer(invalid_param="value")
            @dataclass
            class TestConfig:
                """Test configuration class."""

                field: str


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Type utility functions.
"""

from typing import Union, Any

from ..fallback.typing import get_args, Dict


def type_to_string(type_obj):
    """Converts a Python type to its string representation.

    Args:
        type_obj: Type object to convert.

    Returns:
        str: String representation of the type.

    Note:
        Handles special cases like Union types and types with __origin__.
    """
    if hasattr(type_obj, "__origin__"):
        if type_obj.__origin__ is Union:
            formatted_args = [type_to_string(arg) for arg in get_args(type_obj)]
            return f"Union[{', '.join(formatted_args)}]"
        formatted_args = [type_to_string(arg) for arg in get_args(type_obj)]
        return f"{type_obj.__origin__.__name__}[{', '.join(formatted_args)}]"
    try:
        return type_obj.__name__
    except AttributeError:
        return str(type_obj)


# Alias for typing dict that will be converted to Param dataclass
ParamDict = Dict[str, Any]



#!/usr/bin/env python3
"""
Configuration for custom sources.
"""

from pydantic.dataclasses import dataclass

from .....dataclass.decorators.registry.decorator import registry
from .....dataclass.decorators.exportable.decorator import exportable

from ..base.source import CustomSource
from .base import CustomConfig


@registry(base_cls=CustomSource)
@exportable(strategy="registry")
@dataclass
class CustomSourceConfig(CustomConfig):
    """Custom source configuration."""



#!/usr/bin/env python3
"""
Runner config application export strategy.
"""

import dataclasses
import logging
import pathlib

from pydantic.dataclasses import dataclass

from ..base import ExportStrategy
from ...decorator import exportable
from ...validation import is_exportable
from ......fallback.typing import List, Dict

LOGGER = logging.getLogger(__name__)


@dataclass
class HydraDefaultConfig:
    """Default Hydra configuration."""

    BASE_OUTPUT_DIR: str = "./outputs"  # pylint: disable=C0103
    hydra: Dict = dataclasses.field(
        default_factory=lambda: {
            "run": {"dir": "${BASE_OUTPUT_DIR}/${now:%Y-%m-%d}/${now:%H-%M-%S}"},
            "sweep": {
                "dir": "${BASE_OUTPUT_DIR}/${now:%Y-%m-%d}/${now:%H-%M-%S}",
                "subdir": "run-${hydra:job.num}",
            },
        }
    )


@exportable(stem_file="config_template")
@dataclass
class TemplateDefaultConfigFile(HydraDefaultConfig):
    """Template default config file for a runner application."""

    defaults: List[str] = dataclasses.field(
        default_factory=lambda: [
            {"workflow": "..."},
            {"workenv": "workenv_template"},
            "_self_",
        ]
    )


class RunnerConfigExportStrategy(ExportStrategy, register_name="runner"):
    """Export strategy for runner config applications."""

    @classmethod
    def export(  # pylint: disable=too-many-arguments
        cls,
        data_cls,
        output_dir,
        format_type="yaml",
        exist_ok=False,
        makedirs=False,
        **kwargs,
    ):
        """
        Export the complete runner app configuration structure.
        """

        output_dir = pathlib.Path(output_dir).resolve()

        for field in dataclasses.fields(data_cls):
            field_type = field.type
            field_name = field.name
            if is_exportable(field_type):
                field_type.export(
                    output_dir / field_name,
                    format_type=format_type,
                    exist_ok=exist_ok,
                    makedirs=makedirs,
                    **kwargs,
                )
                continue

            LOGGER.warning(
                "Field %s of type %s is not exportable. Skipping.",
                field_name,
                field_type,
            )

        TemplateDefaultConfigFile.export(  # pylint: disable=no-member
            output_dir, format_type, exist_ok, makedirs, detailed=False
        )



#!/usr/bin/env python3
"""Runner application configuration decorator package."""

from .decorator import runner
from .exceptions import RunnerAttributeError

__all__ = ["runner", "RunnerAttributeError"]



#!/usr/bin/env python3
"""Unit tests for workflow sink components."""

import tempfile
import unittest
from pathlib import Path
from pydantic.dataclasses import dataclass

from pypassist.dataclass.decorators.viewer import viewer
from pypassist.runner.workenv.mixin.sink import SinkMixin
from pypassist.mixin.settings import SettingsMixin
from pypassist.utils.export import export_string
from pypassist.runner.workenv.mixin.exceptions import InvalidConsumeSignatureError
from pypassist.utils.kwargs import validate_kwargs_signature


class TestSink(unittest.TestCase):
    """Test cases for workflow sink components."""

    def setUp(self):
        """Set up test fixtures."""
        self.temp_dir = tempfile.mkdtemp()
        self.output_dir = Path(self.temp_dir)

    def test_basic_sink(self):
        """Test basic sink functionality."""

        @viewer
        @dataclass
        class FileSinkSettings:
            """Settings for file sink."""

            filename: str = "output.txt"
            append: bool = False

        class FileSink(SettingsMixin, SinkMixin):
            """A simple sink that writes data to a file."""

            SETTINGS_DATACLASS = FileSinkSettings

            def __init__(self, settings=None):
                if settings is None:
                    settings = FileSinkSettings()
                SettingsMixin.__init__(self, settings)
                SinkMixin.__init__(self)
                self.consumed_data = None

            def consume(
                self, data=None, export=False, output_dir=None, exist_ok=True, **kwargs
            ):
                """Consume the input data.

                Args:
                    data: The data to consume
                    export: Whether to export the results
                    output_dir: Directory to export results to
                    exist_ok: Whether to overwrite existing files
                    **kwargs: Additional keyword arguments

                Returns:
                    None
                """
                if data is None:
                    raise ValueError("Data parameter is required")

                self.consumed_data = data

                if export and output_dir:
                    # Save the data to a file
                    filepath = output_dir / self.settings.filename
                    mode = "a" if self.settings.append else "w"
                    filepath.parent.mkdir(parents=True, exist_ok=True)

                    with open(filepath, mode, encoding="utf-8") as f:
                        f.write(str(data) + "\n")

                    # Export settings
                    settings_content = self.settings.to_str(format_type="yaml")
                    settings_filepath = (
                        output_dir / f"{self.__class__.__name__.lower()}_settings.txt"
                    )
                    export_string(
                        settings_content,
                        filepath=settings_filepath,
                        exist_ok=exist_ok,
                        makedirs=True,
                    )

        # Test basic consumption
        sink = FileSink()
        sink.consume(data="Test data")
        self.assertEqual(sink.consumed_data, "Test data")

        # Test with custom settings
        settings = FileSinkSettings(filename="custom.txt", append=True)
        sink = FileSink(settings)
        sink.consume(data="Line 1", export=True, output_dir=self.output_dir)
        sink.consume(data="Line 2", export=True, output_dir=self.output_dir)

        # Verify exported files
        output_file = self.output_dir / "custom.txt"
        settings_file = self.output_dir / "filesink_settings.txt"
        self.assertTrue(output_file.exists())
        self.assertTrue(settings_file.exists())
        self.assertEqual(output_file.read_text(), "Line 1\nLine 2\n")

        # Test asset generation
        asset_func = sink.get_assetable_func()
        asset_func(data="Asset data", export=True, output_dir=self.output_dir)
        self.assertEqual(sink.consumed_data, "Asset data")
        self.assertEqual(output_file.read_text(), "Line 1\nLine 2\nAsset data\n")

    def test_consume_signature_validation(self):
        """Test consume signature validation."""

        # Test with valid signature
        class ValidSink(SinkMixin):
            """A sink with a valid consume signature."""

            def __init__(self):
                SinkMixin.__init__(self)
                self.consumed_data = None

            def consume(self, data=None, export=False, output_dir=None, exist_ok=True):
                """Consume with valid signature including all required parameters.

                Args:
                    data: The data to consume
                    export: Whether to export the results
                    output_dir: Directory to export results to
                    exist_ok: Whether to overwrite existing files
                """
                self.consumed_data = data

        # This should work fine
        sink = ValidSink()
        asset_func = sink.get_assetable_func()
        asset_func(
            data="Test data", export=True, output_dir=self.output_dir, exist_ok=True
        )
        self.assertEqual(sink.consumed_data, "Test data")

        # Test with invalid signature
        class InvalidSink:
            """A class with an invalid consume signature."""

            def __init__(self):
                pass

            def consume(self, data):
                """Consume method missing required parameters."""
                pass

        # This should raise InvalidConsumeSignatureError
        invalid_sink = InvalidSink()
        with self.assertRaises(InvalidConsumeSignatureError):
            validate_kwargs_signature(
                obj=invalid_sink,
                method_name="consume",
                required_params=["export", "output_dir", "exist_ok"],
                exception_type=InvalidConsumeSignatureError,
            )

    def test_missing_required_params(self):
        """Test that missing required parameters raises an error."""

        class PartialSink(SinkMixin):
            """A sink with a partially valid signature."""

            def __init__(self):
                SinkMixin.__init__(self)

            def consume(self, data=None, export=False):
                """Consume method missing some required parameters.

                Args:
                    data: The data to consume
                    export: Whether to export the results
                """
                pass

        sink = PartialSink()
        with self.assertRaises(InvalidConsumeSignatureError):
            sink.get_assetable_func()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Registry handlers for type and settings management.
"""

from .exceptions import RegistrySetupError, RegistryAttributeError


class TypeRegistry:
    """Handles type validation and registration."""

    @classmethod
    def validate_type(cls, instance, setup):
        """Validates and normalizes the type.

        Args:
            instance: The instance being processed
            setup: Registry setup configuration

        Returns:
            str: The validated type value

        Raises:
            RegistrySetupError: If the required attribute is missing
            RegistryAttributeError: If the type value is None
        """

        if not hasattr(instance, setup.register_name_attr):
            raise RegistrySetupError(
                f"Missing required attribute '{setup.register_name_attr}'"
            )
        type_value = getattr(instance, setup.register_name_attr)
        if type_value is None:
            raise RegistryAttributeError(f"'{setup.register_name_attr}' cannot be None")
        setattr(instance, setup.register_name_attr, str(type_value))
        return type_value

    @classmethod
    def get_registration(cls, registration_base_class, register_name):
        """Gets the registered class for the type.

        Args:
            registration_base_class: The registration base class
            register_name: The register name for the type to get

        Returns:
            The registered class for the type

        Raises:
            RegistrySetupError: If registration_base_class is not set
        """
        if registration_base_class is None:
            raise RegistrySetupError(
                "registration_base_class must be set in the configuration"
            )
        return registration_base_class.get_registered(register_name)


class SettingsHandler:  # pylint: disable=R0903
    """Handles settings configuration and validation."""

    @classmethod
    def configure_settings(cls, instance, reg_cls, type_value, setup):
        """Configures settings for the instance.

        Args:
            instance: The instance being processed
            reg_cls: The registered class
            type_value: The type value
            setup: Registry setup configuration

        Raises:
            RegistryAttributeError: If the registered class is missing required attributes
            TypeError: If settings value is of invalid type
        """
        if not hasattr(instance, setup.settings_attr):
            return

        settings_value = getattr(instance, setup.settings_attr)
        # pylint : disable=protected-access
        cls._validate_settings_class(reg_cls, setup)
        conf_dcls = getattr(reg_cls, setup.settings_dataclass_attr)
        cls._process_settings(instance, settings_value, conf_dcls, type_value, setup)

    @classmethod
    def _validate_settings_class(cls, reg_cls, setup):
        """Validates that the registered class has the required settings attribute.

        Args:
            reg_cls: The registered class to validate
            setup: Registry setup configuration

        Raises:
            RegistryAttributeError: If the required attribute is missing
        """
        if not hasattr(reg_cls, setup.settings_dataclass_attr):
            raise RegistryAttributeError(
                f"Registered class '{reg_cls.__name__}' missing required "
                f"attribute '{setup.settings_dataclass_attr}'"
            )

    @classmethod
    def _process_settings(  # pylint: disable=too-many-arguments
        cls, instance, settings_value, conf_dcls, type_value, setup
    ):
        """Processes and validates settings value.

        Args:
            instance: The instance being processed
            settings_value: The settings value to process
            conf_dcls: The settings dataclass
            type_value: The type value
            setup: Registry setup configuration

        Raises:
            TypeError: If settings value is of invalid type
        """
        if settings_value is None:
            # Handle None by creating an instance of the expected settings class
            # If it pass validation, it will be set on the instance
            settings_value = conf_dcls()
        if isinstance(settings_value, dict):
            setattr(instance, setup.settings_attr, conf_dcls(**settings_value))
        elif not isinstance(settings_value, conf_dcls):
            raise TypeError(
                f"Invalid type for {type_value} configuration: "
                f"expected {conf_dcls.__name__}, got {type(settings_value).__name__}"
            )



#!/usr/bin/env python3
"""Unit tests for the @runner decorator."""

import unittest
from typing import Dict, Any, List
from pydantic.dataclasses import dataclass

from pypassist.dataclass.decorators.runner import runner, RunnerAttributeError
from pypassist.dataclass.decorators.wenv import wenv
from pypassist.runner.workflow.config.base import WorkflowConfig


class TestRunnerDecorator(unittest.TestCase):
    """Test cases for the @runner decorator functionality."""

    @classmethod
    def setUpClass(cls):
        """Set up test fixtures."""

        @dataclass
        class MinimalWorkflow(WorkflowConfig):
            """Minimal workflow for testing."""

            name: str = "test"
            steps: List[Dict] = None
            output_dir: str = "./output"

            def __post_init__(self):
                if self.steps is None:
                    self.steps = [{"component_path": "test.op"}]
                super().__post_init__()

        cls.MinimalWorkflow = MinimalWorkflow

    def test_basic_decoration(self):
        """Test basic decoration of a dataclass."""

        @wenv
        @dataclass
        class TestWorkEnv:
            """Test work environment."""

            settings: Dict[str, Any] = None

        @runner
        @dataclass
        class TestConfig:
            """Test configuration class."""

            workflow: self.MinimalWorkflow
            workenv: TestWorkEnv

        # Create an instance with valid workflow and workenv
        config = TestConfig(
            workflow=self.MinimalWorkflow(),
            workenv=TestWorkEnv(settings={"param": "value"}),
        )

        # Verify instance creation succeeds
        self.assertIsInstance(config.workflow, WorkflowConfig)
        self.assertTrue(hasattr(config.workenv, "_WENV_"))

    def test_invalid_setup(self):
        """Test handling of invalid setup configuration."""

        # Test missing workflow attribute
        @runner
        @dataclass
        class MissingWorkflowConfig:
            """Test config missing workflow attribute."""

            workenv: Dict[str, Any]

            def __post_init__(self):
                if hasattr(super(), "__post_init__"):
                    super().__post_init__()

        with self.assertRaises(RunnerAttributeError):
            MissingWorkflowConfig(workenv={"settings": {}})

        # Test missing workenv attribute
        @runner
        @dataclass
        class MissingWorkEnvConfig:
            """Test config missing workenv attribute."""

            workflow: self.MinimalWorkflow

            def __post_init__(self):
                if hasattr(super(), "__post_init__"):
                    super().__post_init__()

        with self.assertRaises(RunnerAttributeError):
            MissingWorkEnvConfig(workflow=self.MinimalWorkflow())

        # Test invalid workflow type
        @dataclass
        class InvalidWorkflow:
            """Invalid workflow (not inheriting from WorkflowConfig)."""

            name: str = "test"

        @runner
        @dataclass
        class InvalidWorkflowConfig:
            """Test config with invalid workflow type."""

            workflow: InvalidWorkflow
            workenv: Dict[str, Any]

            def __post_init__(self):
                if hasattr(super(), "__post_init__"):
                    super().__post_init__()

        with self.assertRaises(RunnerAttributeError):
            InvalidWorkflowConfig(
                workflow=InvalidWorkflow(name="test"), workenv={"settings": {}}
            )

        # Test invalid workenv type
        @dataclass
        class InvalidWorkEnv:
            """Invalid work environment (missing @wenv decorator)."""

            settings: Dict[str, Any] = None

        @runner
        @dataclass
        class InvalidWorkEnvConfig:
            """Test config with invalid workenv type."""

            workflow: self.MinimalWorkflow
            workenv: InvalidWorkEnv

            def __post_init__(self):
                if hasattr(super(), "__post_init__"):
                    super().__post_init__()

        with self.assertRaises(RunnerAttributeError):
            InvalidWorkEnvConfig(
                workflow=self.MinimalWorkflow(), workenv=InvalidWorkEnv()
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Processor mixin implementation.
"""

import inspect

from ....utils.kwargs import validate_kwargs_signature
from ..component.processor import ProcessorComponent
from .exceptions import InvalidProcessSignatureError, REQUIRE_KWARGS_SIGNATURE


class ProcessorMixin(ProcessorComponent):
    """Mixin that implements get_assetable_func for processors."""

    def get_assetable_func(self):
        """Convert process method to assetable function.

        Returns:
            A function that can be used as a dagster asset
        """
        validate_kwargs_signature(
            obj=self,
            method_name="process",
            required_params=REQUIRE_KWARGS_SIGNATURE,
            exception_type=InvalidProcessSignatureError,
        )

        def asset_func(**inputs):
            # Simply pass all inputs to process
            return self.process(**inputs)

        # conserve name & signature & type hints
        asset_func.__name__ = self.__class__.__name__
        original_signature = inspect.signature(self.process)
        asset_func.__annotations__ = self.process.__annotations__
        asset_func.__signature__ = original_signature
        return asset_func



#!/usr/bin/env python3
"""Working environment decorator package."""

from .decorator import wenv
from .setup import WenvSetup
from .exceptions import WenvSetupError

__all__ = ["wenv", "WenvSetup", "WenvSetupError"]



#!/usr/bin/env python3
"""
Setup for viewer dataclass decorator.
"""

from dataclasses import dataclass


@dataclass
class ViewerSetup:
    """Setup for viewer decorator."""

    hide_private: bool = True



#!/usr/bin/env python3
"""
Work environment export strategy.
"""

import dataclasses
import pathlib
import logging

from pydantic.dataclasses import dataclass

from ..base import ExportStrategy
from ...validation import is_exportable
from ...decorator import exportable
from ....wenv.validation import is_wenv
from ......fallback.typing import get_args, get_origin, List
from ......utils.typing import type_to_string

LOGGER = logging.getLogger(__name__)

from ......runner.workenv.custom.config.workenv import CustomWorkEnvConfig


class WenvExportStrategy(ExportStrategy, register_name="wenv"):
    """Export strategy for work environment configurations."""

    @classmethod
    def export(  # pylint: disable=too-many-arguments
        cls,
        data_cls,
        output_dir,
        format_type="yaml",
        exist_ok=False,
        makedirs=False,
        **kwargs,
    ):
        """Export the working environment dataclass.

        Args:
            data_cls: The dataclass to export
            output_dir: Directory where to export the data
            format_type: Format to use for export (default: yaml)
            exist_ok: If True, overwrite existing files
            makedirs: If True, create directories if they don't exist
            **kwargs: Additional export options
        """
        output_dir = pathlib.Path(output_dir).resolve()

        # Export each field of the dataclass
        cls._export_fields(
            data_cls, output_dir, format_type, exist_ok, makedirs, **kwargs
        )

        # Export the default configuration
        cls._export_default_workenv(
            data_cls, output_dir, format_type, exist_ok, makedirs, **kwargs
        )

    # -------------------------------------------------------------------------
    # Field export methods
    # -------------------------------------------------------------------------

    @classmethod
    def _export_fields(  # pylint: disable=too-many-arguments
        cls, data_cls, output_dir, format_type, exist_ok, makedirs, **kwargs
    ):
        """Export the fields of the work environment.

        This method handles the export of different field types:
        - Work environment fields
        - Dictionary fields
        - Basic exportable fields
        """
        for field in dataclasses.fields(data_cls):
            field_type = field.type
            if is_wenv(field_type):
                updated_dir = output_dir / field.name
                cls._export_wenv_field(
                    field_type, updated_dir, format_type, exist_ok, makedirs, **kwargs
                )
                continue

            if get_origin(field_type) is dict:
                cls._export_dict_field(
                    field, output_dir, format_type, exist_ok, makedirs, **kwargs
                )
                continue

            if is_exportable(field_type):
                field_type.export(
                    output_dir / field.name,
                    format_type=format_type,
                    exist_ok=exist_ok,
                    makedirs=makedirs,
                    **kwargs,
                )
                continue

            LOGGER.warning(
                "Field '%s' of type '%s' is not exportable. Will be skipped.",
                field.name,
                type_to_string(field_type),
            )

    @classmethod
    def _export_wenv_field(  # pylint: disable=too-many-arguments
        cls, field_type, output_dir, format_type, exist_ok, makedirs, **kwargs
    ):
        """Export a work environment field."""
        field_type.export(
            output_dir,
            format_type=format_type,
            exist_ok=exist_ok,
            makedirs=makedirs,
            **kwargs,
        )

    @classmethod
    def _export_dict_field(  # pylint: disable=too-many-arguments
        cls, field, output_dir, format_type, exist_ok, makedirs, **kwargs
    ):
        """Export a dictionary field."""
        dict_val_cls = get_args(field.type)[1]
        if is_exportable(dict_val_cls):
            field_outdir = output_dir / field.name
            dict_val_cls.export(
                field_outdir,
                format_type=format_type,
                exist_ok=exist_ok,
                makedirs=makedirs,
                **kwargs,
            )

    # -------------------------------------------------------------------------
    # Default configuration export methods
    # -------------------------------------------------------------------------

    @classmethod
    def _export_default_workenv(  # pylint: disable=too-many-arguments
        cls, data_cls, output_dir, format_type, exist_ok, makedirs, **kwargs
    ):
        """Export the default work environment configuration."""
        # Build the list of defaults from the dataclass fields
        defaults = cls._build_defaults_list(data_cls)
        if not defaults:  # Nothing to export
            return
        # Create and export the default configuration
        default_conf = cls._create_default_config(
            defaults, data_cls._EXPORT_SETUP_.stem_file
        )
        default_conf.export(  # pylint: disable=E1101
            output_dir,
            format_type=format_type,
            exist_ok=exist_ok,
            makedirs=makedirs,
            detailed=False,
        )

    @classmethod
    def _build_defaults_list(cls, data_cls):
        """Build the list of defaults from the dataclass fields.

        This method processes each field in the dataclass and builds appropriate
        default entries based on the field type:
        - For dict fields with exportable values: Uses registry or single defaults
        - For basic exportable fields: Uses simple key-value defaults
        """
        defaults = []

        for field in dataclasses.fields(data_cls):
            # Handle dictionary fields (typically for registries)
            if get_origin(field.type) is dict:
                value_type = get_args(field.type)[1]
                if is_exportable(value_type):
                    defaults.extend(cls._build_field_defaults(field.name, value_type))

            # Handle basic exportable fields
            if is_exportable(field.type):
                defaults.extend(
                    cls._build_field_single_basic_default(field.name, field.type)
                )

        return defaults

    @classmethod
    def _build_field_defaults(cls, field_name, value_type):
        """Build defaults for a field based on its type.

        Returns registry defaults if the type has a registry base class,
        otherwise returns a single default entry.
        """
        if hasattr(value_type, "_REG_BASE_CLASS_"):
            return cls._build_registry_defaults(field_name, value_type)
        return [cls._build_single_default(field_name, value_type)]

    @classmethod
    def _build_registry_defaults(cls, field_name, value_type):
        """Build defaults for a registry type.

        Creates a list of defaults for each registered type in the registry,
        using the format: field_name/reg_type@field_name.reg_type
        """
        base_cls = value_type._REG_BASE_CLASS_  # pylint: disable=protected-access
        return [
            f"{field_name}/{reg_type}@{field_name}.{reg_type}"
            for reg_type in base_cls.list_registered()
        ]

    @classmethod
    def _build_single_default(cls, field_name, value_type):
        """Build a single default entry for a field.

        Uses the format: field_name/stem@field_name.stem
        where stem is either the configured stem_file or the lowercase class name.
        """
        stem = (
            value_type._EXPORT_SETUP_.stem_file  # pylint: disable=protected-access
            or value_type.__name__.lower()
        )
        return f"{field_name}/{stem}@{field_name}.{stem}"

    @classmethod
    def _build_field_single_basic_default(cls, field_name, value_type):
        """Build a default entry for a basic exportable field.

        Returns a list containing a single dictionary with the field name as key
        and stem as value, which will be properly serialized in YAML format.
        """
        stem = (
            value_type._EXPORT_SETUP_.stem_file  # pylint: disable=protected-access
            or value_type.__name__.lower()
        )
        return [{field_name: stem}]

    @classmethod
    def _create_default_config(cls, defaults, stem_file=None):
        """Create the default configuration dataclass.

        Args:
            defaults: List of default entries to include in the configuration

        Returns:
            A dataclass configured with the provided defaults
        """
        if stem_file is None:
            stem_file = "workenv_template"

        @exportable(stem_file=stem_file)
        @dataclass
        class DefaultConfig:  # pylint: disable=C0115
            defaults: List = dataclasses.field(default_factory=lambda: defaults)

        return DefaultConfig



#!/usr/bin/env python3
"""Unit tests for the @registry decorator."""

import unittest
from typing import Dict, Any
from pydantic.dataclasses import dataclass
from abc import ABC, abstractmethod

from pypassist.mixin.registrable import Registrable
from pypassist.dataclass.decorators.registry import registry, RegistrySetupError
from pypassist.dataclass.decorators.registry.setup import RegistrySetup


class TestRegistryDecorator(unittest.TestCase):
    """Test cases for the @registry decorator functionality."""

    def setUp(self):
        """Set up test fixtures."""

        class BaseProcessor(ABC, Registrable):
            """Base class for testing registry."""

            _REGISTER = {}

            @abstractmethod
            def process(self, data: str) -> str:
                """Process the data."""

    def test_basic_decoration(self):
        """Test basic decoration of a dataclass."""

        class BaseProcessor(ABC, Registrable):
            """Base class for testing registry."""

            _REGISTER = {}

            @abstractmethod
            def process(self, data: str) -> str:
                """Process the data."""

        @registry(base_cls=BaseProcessor)
        @dataclass
        class ProcessorConfig:
            """Test configuration class."""

            type: str = "test"
            settings: Dict[str, Any] = None

        # Create an instance
        config = ProcessorConfig(type="test", settings={"param": "value"})

        # Verify the class is properly decorated
        self.assertTrue(hasattr(ProcessorConfig, "_SETUP_"))
        self.assertTrue(hasattr(ProcessorConfig, "_REG_BASE_CLASS_"))
        self.assertTrue(hasattr(ProcessorConfig, "_REGISTRY_"))
        self.assertTrue(hasattr(ProcessorConfig, "reload_registry"))

    def test_invalid_setup(self):
        """Test handling of invalid setup configuration."""

        class BaseProcessor(ABC, Registrable):
            """Base class for testing registry."""

            _REGISTER = {}

            @abstractmethod
            def process(self, data: str) -> str:
                """Process the data."""

        with self.assertRaises(RegistrySetupError):

            @registry(
                base_cls=BaseProcessor, register_name_attr=123
            )  # Should be string
            @dataclass
            class ProcessorConfig:
                """Test configuration class."""

                type: str = "test"
                settings: Dict[str, Any] = None


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Utilities for working with Hydra.
"""

import os
from pathlib import Path


def get_working_dir():
    """Get working directory where the app was launched from.

    Returns:
        Path: Working directory path
    """
    # Get absolute path of the current working directory
    return Path(os.getcwd()).resolve()


def get_config_path():
    """Get path to config directory relative to working directory.

    Returns:
        Path: Path to config directory
    """
    working_dir = get_working_dir()
    return working_dir / "config"



#!/usr/bin/env python3
"""
YAML formatter.
"""

from collections import OrderedDict
import dataclasses
import logging

import textwrap
import yaml
from pydantic.dataclasses import dataclass

from ..base import Formatter
from ..info import DataclassInfo
from ....utils.ipynb import is_in_ipynb
from ....utils.convert import to_dict_recursive

LOGGER = logging.getLogger(__name__)


@dataclass
class YamlFormatterSettings:
    """YAML formatter settings."""

    max_width: int = 80
    indent: str = "  "
    indent_level: int = 0
    show_header: bool = False  # whether to show header
    detailed: bool = True  # whether to use detailed mode
    show_docs: bool = True  # whether to show docs in detailed mode


class YamlFormatter(Formatter, register_name="yaml"):
    """YAML formatter."""

    SETTINGS_DATACLASS = YamlFormatterSettings

    def __init__(self, settings=None):
        if settings is None:
            settings = YamlFormatterSettings()
        super().__init__(settings)

    def to_serialized(self, class_info):
        """Convert DataclassInfo to a serialized object."""
        return self.to_str(class_info)

    def to_str(self, class_info):
        """Format DataclassInfo into YAML string."""
        if self.settings.detailed:
            return self._format_detailed(class_info)
        return self._format_simple(class_info)

    def display(self, content):
        """Display the formatted content."""
        if is_in_ipynb():
            # pylint: disable=import-outside-toplevel
            from IPython.display import display_markdown

            LOGGER.debug("Displaying content in Jupyter Notebook.")
            display_markdown(f"```yaml\n{content}\n```", raw=True)
        else:
            LOGGER.debug("Displaying content in terminal.")
            print(content)

    def _format_detailed(self, class_info):
        """Format class info in detailed mode."""
        content = []

        if self.settings.show_header:
            content.extend(self._format_header(class_info))

        if class_info.is_empty:
            content.append(class_info.empty_message)
        else:
            content.extend(self._format_detailed_fields(class_info))

        return "\n".join(content)

    def _format_detailed_fields(self, class_info):
        """Format the fields of a non-empty dataclass in detailed mode."""
        content = []
        current_indent = self.settings.indent * self.settings.indent_level

        for field in class_info.fields:
            if self.settings.show_docs:
                if field.description:
                    content.extend(self._wrap_comment(field.description))
                content.append(f"{current_indent}## Type: {field.type_info}")

            if isinstance(field.value, DataclassInfo):
                nested_settings = dataclasses.replace(
                    self.settings, indent_level=self.settings.indent_level + 1
                )
                nested_formatter = YamlFormatter(settings=nested_settings)
                content.append(f"{current_indent}{field.name}:")
                nested_content = nested_formatter.to_str(field.value)
                content.extend(
                    f"{current_indent}{self.settings.indent}{line}"
                    for line in nested_content.splitlines()
                )
            else:
                value_str = yaml.safe_dump(
                    {field.name: field.value}, **self._yaml_dump_kwargs
                ).rstrip()
                content.extend(
                    f"{current_indent}{line}" for line in value_str.splitlines()
                )

        return content

    def _format_simple(self, class_info):
        """Format class info in simple mode (attributes = values)."""
        if class_info.is_empty:
            return class_info.empty_message

        data = self._build_simple(class_info)
        data_dict = to_dict_recursive(data)  # OrderedDict -> dict
        return yaml.safe_dump(data_dict, **self._yaml_dump_kwargs)

    def _build_simple(self, class_info):
        """Build a simple dictionary representation."""
        result = OrderedDict()
        for field_info in class_info.fields:
            if isinstance(field_info.value, DataclassInfo):
                result[field_info.name] = self._build_simple(field_info.value)
            else:
                result[field_info.name] = field_info.value
        return result

    def _format_header(self, class_info):
        """Format the header section with class name and description."""
        result = []
        base_indent = self.settings.indent * self.settings.indent_level
        header_width = self.settings.max_width // 2

        result.append(f"{base_indent}# {'~'*header_width}")
        result.append(f"{base_indent}# {class_info.name.upper()}")

        if self.settings.show_docs and class_info.description:
            result.extend(self._wrap_comment(class_info.description))

        result.append(f"{base_indent}# {'~'*header_width}")
        return result

    def _wrap_comment(self, text):
        """Wrap text into YAML comments."""
        indent_level = self.settings.indent_level
        indent = self.settings.indent * indent_level
        wrapped = textwrap.wrap(
            text,
            width=self.settings.max_width - len(indent) - 2,
            break_long_words=False,
            break_on_hyphens=False,
        )
        return [f"{indent}## {line}" for line in wrapped]

    @property
    def _yaml_dump_kwargs(self):
        """Common yaml.safe_dump parameters."""
        return {
            "default_flow_style": False,
            "indent": len(self.settings.indent),
            "width": self.settings.max_width,
            "allow_unicode": True,
        }



#!/usr/bin/env python3
"""
Base export strategy.
"""

from abc import ABC, abstractmethod

from .....mixin.registrable import Registrable


class ExportStrategy(ABC, Registrable):
    """Base export strategy."""

    _REGISTER = {}
    _TYPE_SUBMODULE = "type"

    # pylint: disable=R0913
    @classmethod
    @abstractmethod
    def export(
        cls,
        data_cls,
        output_dir,
        format_type="yaml",
        exist_ok=False,
        makedirs=True,
        **kwargs,
    ):
        """Export the configuration.

        Args:
            cls: The class to export
            output_dir: Output directory
            format_type: Output format (yaml, json, etc.)
            exist_ok: Overwrite existing files
            makedirs: Create missing directories
            **kwargs: Additional formatting options
        """



#!/usr/bin/env python3
"""
Registrable mixin module for automatic class registration capabilities.
"""

from .mixin import Registrable
from .exceptions import UnregisteredTypeError

__all__ = [
    "Registrable",
    "UnregisteredTypeError",
]



#!/usr/bin/env python3
"""
Exceptions for the registrable mixin.
"""

from ...exceptions import PypassistError


class RegistryError(PypassistError):
    """Base exception for registry-related errors."""


class InvalidRegistrationNameError(RegistryError):
    """Raised when a registration name is invalid."""


class RegistryImportError(RegistryError):
    """Raised when there is an error importing subtypes from a module."""


class UnregisteredTypeError(RegistryError):
    """Raised when a requested registration name is not found in the registry."""



#!/usr/bin/env python3
"""
Runner app configuration exceptions.
"""

from ...exceptions import DataclassError


class RunnerError(DataclassError):
    """Base exception for runner application configuration related errors."""


class RunnerAttributeError(RunnerError):
    """Exception for runner application configuration attribute errors."""



#!/usr/bin/env python3
"""Helper mixin to convert strings to enum values."""


import json
import yaml


class EnumStrMixin:
    """
    Mixin class that provides a classmethod to parse a string into an Enum.
    and a property to get the value as a string.
    """

    def __init_subclass__(cls, **kwargs):
        """Called when a class inherits from this mixin."""
        super().__init_subclass__(**kwargs)

        def enum_representer(dumper, data):
            """Use __str__ for YAML representation."""
            return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))

        yaml.add_representer(cls, enum_representer, Dumper=yaml.SafeDumper)

        json.JSONEncoder.default = lambda self, data: (
            str(data) if isinstance(data, cls) else json.JSONEncoder.default(self, data)
        )

    @classmethod
    def from_str(cls, value, info=None):  # pylint: disable=unused-argument
        """Parse a string into an Enum value.

        Handles various string formats intelligently:
        - Case-insensitive matching
        - Strips whitespace

        Args:
            value (Any): The string to parse or an existing enum value.
            info (pydantic.ValidationInfo, optional): Additional validation metadata
                provided by Pydantic. This parameter is ignored but included for compatibility.

        Returns:
            EnumStrMixin: An instance of the Enum.

        Raises:
            ValueError: If the string cannot be converted to a valid enum value.
            TypeError: If the input is not a string or an instance of the Enum.
        """
        if isinstance(value, cls):
            return value

        if not isinstance(value, str):
            raise ValueError(
                f"Expected string or {cls.__name__}, " f"got {type(value).__name__}"
            )

        # Normalize input
        value = value.strip().upper()

        try:
            return cls[value]
        except KeyError as err:
            valid_values = ", ".join(f"'{v.name.lower()}'" for v in cls)
            raise ValueError(
                f"Invalid {cls.__name__.lower()}: '{value}'. "
                f"Valid values are: {valid_values}"
            ) from err

    @property
    def label(self):
        """
        Returns a lowercase, human-readable version of the enum name.
        """
        return self.name.upper()

    def upper(self):
        """
        Returns a uppercase, human-readable version of the enum name.
        """
        return self.name.upper()

    def lower(self):
        """
        Returns a lowercase, human-readable version of the enum name.
        """
        return self.name.lower()

    def __str__(self):
        """Returns string representation for serialization."""
        return self.label

    @classmethod
    def __get_validators__(cls):
        """Pydantic hook to use `from_str` automatically during validation."""
        yield cls.from_str



#!/usr/bin/env python3
"""
Working environment decorator.
"""

import logging

from ....fallback.typing import get_origin
from ..base import BaseDecorator
from .exceptions import WenvSetupError
from .setup import WenvSetup
from .dict_field import DictFieldProcessor

LOGGER = logging.getLogger(__name__)


def wenv(_cls=None, *, setup=None, **kwargs):
    """Work environment configuration decorator.

    This decorator ensures proper initialization of nested dictionary fields,
    converting dict values to their corresponding dataclasses.
    """

    def wrap(cls):
        try:
            cls, setup_instance = BaseDecorator.decorate_class(
                cls, WenvSetup, setup, **kwargs
            )
        except ValueError as err:
            raise WenvSetupError(str(err)) from err

        class WenvClass(cls):  # pylint: disable=R0903, C0115
            cls._SETUP_ = setup_instance
            cls._WENV_ = True

            def __post_init__(self):
                if hasattr(super(), "__post_init__"):  # parent post-init
                    super().__post_init__()

                for field in self.__class__.__dataclass_fields__.values():
                    if get_origin(field.type) is dict:
                        DictFieldProcessor.process_field(self, field)

        BaseDecorator.preserve_cls_metadata(WenvClass, cls)
        return WenvClass

    if _cls is None:
        return wrap
    return wrap(_cls)



#!/usr/bin/env python3
"""
Base decorator for dataclass decorators.
"""

import dataclasses
import logging

from pydantic.dataclasses import dataclass

from ...fallback.pydantic import is_pydantic_dataclass

LOGGER = logging.getLogger(__name__)


class BaseDecorator:
    """Base decorator for dataclass decorators."""

    @staticmethod
    def decorate_class(decorated_cls, setup_cls, setup=None, **kwargs):
        """Base wrapper for dataclass decorators.

        Args:
            decorated_cls: Class to decorate
            setup_cls: Setup class to use (e.g. ViewerSetup, RegistrySetup)
            setup: Setup instance or dict
            kwargs: Additional setup parameters

        Returns:
            tuple: (processed_class, setup_instance)

        Raises:
            ValueError: If setup is not of the expected type
        """
        # Setup handling
        if setup is None:
            setup = setup_cls()
        if isinstance(setup, dict):
            setup = setup_cls(**setup)
        if not isinstance(setup, setup_cls):
            raise ValueError(
                f"Expected {setup_cls.__name__} but got {type(setup).__name__}"
            )
        if kwargs:
            setup = dataclasses.replace(setup, **kwargs)

        # Ensure pydantic dataclass
        if not is_pydantic_dataclass(decorated_cls):
            LOGGER.warning(
                "Decorator requires pydantic dataclass. "
                "Converting %s to pydantic dataclass. "
                "Use pydantic_dataclass decorator instead to suppress this warning.",
                decorated_cls.__name__,
            )
            decorated_cls = dataclass(decorated_cls)

        return decorated_cls, setup

    @staticmethod
    def preserve_cls_metadata(wrapped_cls, original_cls):
        """Preserve original class metadata.

        Args:
            wrapped_cls: The wrapped class
            original_cls: The original class
        """
        if "__pydantic_validator__" in original_cls.__dict__:
            wrapped_cls.__pydantic_validator__ = original_cls.__pydantic_validator__

        wrapped_cls.__name__ = original_cls.__name__
        wrapped_cls.__qualname__ = original_cls.__qualname__
        wrapped_cls.__module__ = original_cls.__module__



#!/usr/bin/env python3
"""
Common exceptions for dataclass decorators.
"""

from ..exceptions import PypassistError


class DataclassError(PypassistError):
    """Base exception for all dataclass-related errors."""


class FormattingError(DataclassError):
    """Exception for formatting-related errors."""



#!/usr/bin/env python3
"""
Common mixin to handle settings.
"""

import dataclasses

import logging

from ..utils.export import export_string
from ..fallback.typing import NoneType

LOGGER = logging.getLogger(__name__)


class SettingsMixin:
    """
    Enhanced mixin for handling settings with nested updates support.

    Attributes:
        SETTINGS_DATACLASS: The dataclass to be used for settings.
    """

    SETTINGS_DATACLASS = NoneType

    def __init__(self, settings):
        settings = self._check_settings_type(settings)
        self._settings = settings

    @property
    def settings(self):
        """Get the current settings."""
        if self.SETTINGS_DATACLASS is NoneType:
            LOGGER.warning(
                "%s does not have settings. Returning None", self.__class__.__name__
            )
            return None
        return self._settings

    def _replace_settings_if_not_none(self, new_settings):
        """Replace settings if the new settings are not None."""
        if new_settings is not None:
            self._check_settings_type(new_settings)
            self._settings = dataclasses.replace(new_settings)

    def update_settings(self, settings=None, **kwargs):
        """
        Update the settings either with a new settings object or individual values.

        Args:
            settings: New settings object to update from
            **kwargs: Individual settings to update using dot notation for nested attributes
        """
        if self.SETTINGS_DATACLASS is NoneType:
            LOGGER.warning(
                "%s does not have settings. Ignoring update_settings call",
                self.__class__.__name__,
            )
            return

        self._replace_settings_if_not_none(settings)

        if not kwargs:
            return

        updates = {}
        nested_updates = {}

        for key, value in kwargs.items():
            if "." in key:
                nested_updates[key] = value
            else:
                updates[key] = value

        if updates:
            try:
                self._settings = dataclasses.replace(self._settings, **updates)
            except (AttributeError, TypeError) as err:
                raise ValueError(f"Invalid settings update: {err}") from err

        for path, value in nested_updates.items():
            try:
                self._update_nested_attribute(path, value)
            except AttributeError as err:
                LOGGER.warning("Failed to update nested setting `%s`: %s", path, err)

        self._trigger_post_init()

    def export_settings(
        self, output_dir, format_type="yaml", exist_ok=False, makedirs=False
    ):
        """
        Export the settings to a file.

        Args:
            output_dir: Directory where to export the settings
            format_type: Format to use for export (default: yaml)
            exist_ok: Whether to overwrite existing files (default: False)
            makedirs: Whether to create directories if they don't exist (default: False)
        """
        if self.SETTINGS_DATACLASS is NoneType:
            LOGGER.info(
                "%s does not have settings. Skipping settings export.",
                self.__class__.__name__,
            )
            return

        content = self.settings.to_str(
            format_type=format_type,
            output_dir=output_dir,
            detailed=False,
            exist_ok=exist_ok,
        )
        filepath = output_dir / f"{self.__class__.__name__.lower()}_settings.txt"
        export_string(content, filepath=filepath, exist_ok=exist_ok, makedirs=makedirs)

    def _update_nested_attribute(self, path, value):
        """
        Update a nested attribute using dot notation.

        Args:
            path: Dot-separated path to the attribute (e.g., "optimizer.params.lr")
            value: New value to set
        """
        attrs = path.split(".")
        current = self._settings

        for attr in attrs[:-1]:
            if not hasattr(current, attr):
                raise AttributeError(f"Invalid nested attribute `{attr}` in `{path}`")
            current = getattr(current, attr)

        last_attr = attrs[-1]
        if not hasattr(current, last_attr):
            raise AttributeError(f"Invalid final attribute `{last_attr}` in `{path}`")

        if dataclasses.is_dataclass(current):
            setattr(current, last_attr, value)
            current = dataclasses.replace(current, **{last_attr: value})
        else:
            setattr(current, last_attr, value)

    def _check_settings_type(self, settings=None):
        """
        Validate settings type and convert dict to proper settings instance if needed.
        """
        # Handle NoneType class (no settings)
        if self.SETTINGS_DATACLASS is NoneType:
            LOGGER.info(
                "%s does not support settings. Returning None.", self.__class__.__name__
            )
            return None

        if isinstance(settings, dict):
            try:
                settings = self.SETTINGS_DATACLASS(**settings)  # pylint: disable=E1102
            except (TypeError, ValueError) as err:
                raise TypeError(f"Invalid settings dictionary: {err}") from err

        if not isinstance(settings, self.SETTINGS_DATACLASS):  # pylint: disable=W1116
            raise TypeError(
                f"Invalid settings type: {type(settings).__name__}. "
                f"Expected: {self.SETTINGS_DATACLASS.__name__}"
            )

        return settings

    def _trigger_post_init(self):
        """Trigger post_init if available on the settings."""
        if hasattr(self._settings, "__post_init__"):
            self._settings.__post_init__()



#!/usr/bin/env python3
"""
Registry exceptions.
"""

from ...exceptions import DataclassError


class RegistryError(DataclassError):
    """Base exception for all Registry related errors."""


class RegistrySetupError(RegistryError):
    """Raised when there is an error in the Registry setup configuration."""


class RegistrationErrror(RegistryError):
    """Raised when there is an error during type registration."""


class RegistryAttributeError(RegistryError):
    """Raised when there is an error during validation of attributes."""



#!/usr/bin/env python3
"""
Dataclass validation.
"""

from .setup import WenvSetup


def is_wenv(obj):
    """
    Return True if the given object is a wenv dataclass or an instance
    of a wenv dataclass.
    """
    if isinstance(obj, type):
        dtcls = obj
    else:
        dtcls = type(obj)
    if hasattr(dtcls, "_SETUP_"):
        setup = getattr(dtcls, "_SETUP_")
        if isinstance(setup, WenvSetup):
            return True
    return False



#!/usr/bin/env python3
"""
Component exceptions
"""

from ...exceptions import RunnerError


class ComponentError(RunnerError):
    """Base component error."""


class InvalidComponentError(ComponentError):
    """Invalid component."""



#!/usr/bin/env python3
"""
Base execution concepts for workflow module.
"""

from abc import ABC, abstractmethod


class BaseStep(ABC):
    """Base class for workflow steps."""

    def __init__(self, config, workenv):
        self._config = config
        self.workenv = workenv
        self._component = self._config.get_component(self.workenv)

    @property
    def config(self):
        """Return step configuration."""
        return self._config

    @property
    def component(self):
        """Return step component."""
        return self._component

    @property
    def inputs(self):
        """Return step inputs."""
        return self.config.inputs

    @property
    def output_dir(self):
        """Return Path to output directory."""
        return self.config.output_dir

    @property
    def exist_ok(self):
        """Return exist_ok flag."""
        return self.config.exist_ok

    @property
    def save_results(self):
        """Return save_results flag."""
        return self.config.save_results

    @abstractmethod
    def execute(self, **inputs):
        """Execute the step."""


class BaseWorkflow(ABC):
    """Base class for workflows."""

    def __init__(self, config, workenv):
        self.config = config
        self._steps = []
        self._initialize_steps(workenv)

    @abstractmethod
    def _initialize_steps(self, workenv):
        """Initialize workflow steps."""

    @property
    def steps(self):
        """Return workflow steps."""
        return self._steps

    @property
    def components(self):
        """Return workflow components."""
        return [step.component for step in self.steps]

    @abstractmethod
    def execute(self):
        """Execute the workflow."""



#!/usr/bin/env python3
"""
Exportable validation.
"""

from .setup import ExportableSetup


def is_exportable(obj):
    """
    Return True if the given object is a exportable dataclass or an instance
    of a exportable dataclass.
    """
    if isinstance(obj, type):
        dtcls = obj
    else:
        dtcls = type(obj)

    if hasattr(dtcls, "_EXPORT_SETUP_"):
        setup = getattr(dtcls, "_EXPORT_SETUP_")
        if isinstance(setup, ExportableSetup):
            return True
    return False



#!/usr/bin/env python3
"""
Configuration for custom components.
"""

import dataclasses
from typing import Optional

from .....utils.typing import ParamDict


@dataclasses.dataclass
class CustomConfig:
    """Configuration for custom components."""

    name: str
    settings: Optional[ParamDict] = None

    def get_component(self, workenv=None):
        """Get processor instance.

        Args:
            workenv: Optional work environment instance

        Returns:
            An instance of the processor
        """
        # pylint: disable=no-member
        return self._REG_BASE_CLASS_.get_registered(self.name).init_from_config(
            self, workenv=workenv
        )



#!/usr/bin/env python3
"""
Base class for custom source components.
"""

from abc import abstractmethod

from .component import CustomComponent
from ...mixin.source import SourceMixin


class CustomSource(CustomComponent, SourceMixin):
    """Base class for custom source components.

    Users should extend this class to create their own custom sources.
    """

    _REGISTER = {}
    _TYPE_SUBMODULE = None

    def __init__(self, settings, *, workenv=None):
        """Initialize the source.

        Args:
            settings: Source settings
        """
        CustomComponent.__init__(self, settings, workenv=workenv)
        SourceMixin.__init__(self)

    @abstractmethod
    def provide(self, **kwargs):
        """Provide with flexible signature.

        Args:
            **kwargs: Arbitrary keyword arguments including:
                - export: Whether to export the results
                - output_dir: Directory to export results to
                - exist_ok: Whether to overwrite existing files

        Returns:
            The provided data
        """



#!/usr/bin/env python3
"""
Base class for custom sink components.
"""

from abc import abstractmethod

from .component import CustomComponent
from ...mixin.sink import SinkMixin


class CustomSink(CustomComponent, SinkMixin):
    """Base class for custom sink components.

    Users should extend this class to create their own custom sinks.
    """

    _REGISTER = {}
    _TYPE_SUBMODULE = None

    def __init__(self, settings, *, workenv=None):
        """Initialize the sink.

        Args:
            settings: Sink settings
        """
        CustomComponent.__init__(self, settings, workenv=workenv)
        SinkMixin.__init__(self)

    @abstractmethod
    def consume(self, **kwargs):
        """Consume with flexible signature.

        Args:
            **kwargs: Arbitrary keyword arguments including:
                - export: Whether to export the results
                - output_dir: Directory to export results to
                - exist_ok: Whether to overwrite existing files

        Returns:
            The result of consuming the data
        """



#!/usr/bin/env python3
"""
Viewer validation.
"""

from .setup import ViewerSetup


def is_viewer(obj):
    """
    Return True if the given object is a viewer dataclass or an instance
    of a viewer dataclass.
    """
    dataclass_type = type(obj)
    if hasattr(dataclass_type, "_SETUP_"):
        setup = getattr(dataclass_type, "_SETUP_")
        if isinstance(setup, ViewerSetup):
            return True
    return False



#!/usr/bin/env python3
"""
Dataclass information classes supporting both instances and schema definitions.
"""

import dataclasses
import inspect
from typing import Any
from docstring_parser import parse as parse_docstring

from ...utils.typing import type_to_string
from ...fallback.typing import List, NoneType
from ...fallback.pydantic import is_pydantic_dataclass

# from ...utils.empty_settings import NoneType
from ..exceptions import DataclassError


def _get_extra_metadata(field):
    """Get field metadata if any."""
    return getattr(field, "metadata", None)


def is_metadata_required(field, required_flag="_FORCE_REQUIRED_"):
    """Check if field is required through metadata override."""
    extra = _get_extra_metadata(field)
    return extra is not None and extra.get(required_flag, False) is True


def is_default_missing(field):
    """Check if field has no default value."""
    return (
        field.default is dataclasses.MISSING
        and field.default_factory is dataclasses.MISSING
    )


def is_required_field(field):
    """Check if field should be treated as required."""
    return is_metadata_required(field) or is_default_missing(field)


@dataclasses.dataclass
class DataclassFieldInfo:
    """Information about a dataclass field."""

    name: str
    type_info: str
    description: str
    value: Any

    @classmethod
    def from_field(cls, field, instance=None, docs=None):
        """Create DataclassFieldInfo from a dataclass Field."""
        type_string = type_to_string(field.type)
        field_description = docs.get(field.name) if docs else "UNDOCUMENTED"

        required_field = is_required_field(field)

        if instance is not None:
            # Instance mode
            field_value = getattr(instance, field.name)

            if field_value is None and not required_field:
                field_value = field.default

            if field_value is not None and is_pydantic_dataclass(type(field_value)):
                field_value = DataclassInfo.from_instance(field_value)
        else:
            # Schema mode
            if not required_field or is_metadata_required(field):
                if field.default is not dataclasses.MISSING:
                    field_value = field.default
                else:
                    field_value = field.default_factory()
            else:
                field_value = "..."

            if is_pydantic_dataclass(field.type):
                field_value = DataclassInfo.from_class(field.type)

        required_status = " [REQUIRED]" if required_field else " [OPTIONAL]"
        type_info = f"{type_string}{required_status}"

        return cls(
            name=field.name,
            type_info=type_info,
            description=field_description,
            value=field_value,
        )


@dataclasses.dataclass
class DataclassInfo:
    """Information about a dataclass."""

    name: str
    description: str
    fields: List[DataclassFieldInfo] = None
    is_empty: bool = dataclasses.field(init=False, default=False)
    empty_message: str = dataclasses.field(init=False, default="")

    _DEFAULT_EMPTY_MESSAGE = "## No field defined in: {name}"

    def __post_init__(self):
        if self.fields is None:
            self.fields = []

        # is dataclass empty
        self.is_empty = len(self.fields) == 0
        if self.is_empty:
            self.empty_message = (
                f"\n{self._DEFAULT_EMPTY_MESSAGE.format(name=self.name)}"
            )

    @staticmethod
    def parse_docstring(docstring):
        """Parse docstring using docstring_parser."""
        if not docstring:
            return "", {}

        parsed = parse_docstring(docstring)
        description_parts = []

        if parsed.short_description:
            description_parts.append(parsed.short_description)
        if parsed.long_description:
            description_parts.append(parsed.long_description)

        description = "\n\n".join(description_parts)
        attr_docs = {param.arg_name: param.description for param in parsed.params}

        description = description.strip() or "UNDOCUMENTED"
        return description, attr_docs

    @classmethod
    def from_instance(cls, instance, hide_private=True):
        """Create DataclassInfo from a dataclass instance."""
        dataclass_type = instance.__class__
        cls._check_type(dataclass_type)
        docstring = inspect.getdoc(dataclass_type)
        class_desc, attr_docs = cls.parse_docstring(docstring)

        dc_fields = [
            f
            for f in dataclasses.fields(instance)
            if not (hide_private and f.name.startswith("_"))
        ]

        field_infos = [
            DataclassFieldInfo.from_field(field=f, instance=instance, docs=attr_docs)
            for f in dc_fields
        ]

        return cls(
            name=dataclass_type.__name__,
            description=class_desc,
            fields=field_infos,
        )

    @classmethod
    def from_class(cls, dataclass_type, hide_private=True):
        """Create DataclassInfo from a dataclass type for schema definition."""
        cls._check_type(dataclass_type)

        docstring = inspect.getdoc(dataclass_type)
        class_desc, attr_docs = cls.parse_docstring(docstring)

        dc_fields = [
            f
            for f in dataclasses.fields(dataclass_type)
            if not (hide_private and f.name.startswith("_"))
        ]

        field_infos = [
            DataclassFieldInfo.from_field(field=f, docs=attr_docs) for f in dc_fields
        ]

        return cls(
            name=dataclass_type.__name__,
            description=class_desc,
            fields=field_infos,
        )

    @classmethod
    def _check_type(cls, dataclass_type):
        validate_pydantic_dataclass(dataclass_type)


def validate_pydantic_dataclass(dataclass_type):
    """Check if the given object is a pydantic dataclass or NoneType."""
    # Special case for NoneType
    if dataclass_type is NoneType:
        return

    if not is_pydantic_dataclass(dataclass_type):
        raise DataclassError(
            f"Expected pydantic dataclass, got {dataclass_type.__name__}"
        )


def has_required_fields(datacls):
    """
    Check if at least one fields in a dataclass are required.

    A field is considered required if:
    - It has no default value and no default_factory
    - OR if its metadata contains '_FORCE_REQUIRED_': True

    Args:
        datacls: The dataclass to check

    Returns:
        bool: True if all fields are required, False otherwise
    """
    # Special case for NoneType
    if datacls is NoneType:
        return False

    validate_pydantic_dataclass(datacls)
    fields = dataclasses.fields(datacls)
    return any(is_required_field(field) for field in fields)



#!/usr/bin/env python3
"""Unit tests for workflow processor components."""

import tempfile
import unittest
from pathlib import Path
from pydantic.dataclasses import dataclass

from pypassist.dataclass.decorators.viewer import viewer
from pypassist.runner.workenv.mixin.processor import ProcessorMixin
from pypassist.mixin.settings import SettingsMixin
from pypassist.utils.export import export_string
from pypassist.runner.workenv.mixin.exceptions import InvalidProcessSignatureError
from pypassist.utils.kwargs import validate_kwargs_signature


class TestProcessor(unittest.TestCase):
    """Test cases for workflow processor components."""

    def setUp(self):
        """Set up test fixtures."""
        self.temp_dir = tempfile.mkdtemp()
        self.output_dir = Path(self.temp_dir)

    def test_basic_processor(self):
        """Test basic processor functionality."""

        @viewer
        @dataclass
        class ProcessorSettings:
            """Settings for test processor."""

            multiplier: int = 2

        class BasicProcessor(SettingsMixin, ProcessorMixin):
            """A simple processor that multiplies input by a factor."""

            SETTINGS_DATACLASS = ProcessorSettings

            def __init__(self, settings=None):
                if settings is None:
                    settings = ProcessorSettings()
                SettingsMixin.__init__(self, settings)
                ProcessorMixin.__init__(self)

            def process(
                self, value=None, export=False, output_dir=None, exist_ok=True, **kwargs
            ):
                """Process the input value.

                Args:
                    value: The value to process
                    export: Whether to export the results
                    output_dir: Directory to export results to
                    exist_ok: Whether to overwrite existing files
                    **kwargs: Additional keyword arguments

                Returns:
                    The processed value
                """
                if value is None:
                    raise ValueError("Value parameter is required")

                result = value * self.settings.multiplier

                if export and output_dir:
                    filepath = (
                        output_dir / f"{self.__class__.__name__.lower()}_result.txt"
                    )
                    export_string(
                        str(result), filepath=filepath, exist_ok=exist_ok, makedirs=True
                    )

                    settings_content = self.settings.to_str(format_type="yaml")
                    settings_filepath = (
                        output_dir / f"{self.__class__.__name__.lower()}_settings.txt"
                    )
                    export_string(
                        settings_content,
                        filepath=settings_filepath,
                        exist_ok=exist_ok,
                        makedirs=True,
                    )

                return result

        # Test basic processing
        processor = BasicProcessor()
        self.assertEqual(processor.process(value=5), 10)

        # Test with custom settings
        settings = ProcessorSettings(multiplier=3)
        processor = BasicProcessor(settings)
        self.assertEqual(processor.process(value=5), 15)

        # Test asset generation
        asset_func = processor.get_assetable_func()
        result = asset_func(value=5, export=True, output_dir=self.output_dir)
        self.assertEqual(result, 15)

        # Verify exported files
        result_file = self.output_dir / "basicprocessor_result.txt"
        settings_file = self.output_dir / "basicprocessor_settings.txt"
        self.assertTrue(result_file.exists())
        self.assertTrue(settings_file.exists())
        self.assertEqual(result_file.read_text(), "15")

    def test_process_signature_validation(self):
        """Test process signature validation."""

        # Test with valid signature
        class ValidProcessor(ProcessorMixin):
            """A processor with a valid process signature."""

            def __init__(self):
                ProcessorMixin.__init__(self)

            def process(self, value=None, export=False, output_dir=None, exist_ok=True):
                """Process with valid signature including all required parameters.

                Args:
                    value: The value to process
                    export: Whether to export the results
                    output_dir: Directory to export results to
                    exist_ok: Whether to overwrite existing files

                Returns:
                    The processed value
                """
                return value * 2

        # This should work fine
        processor = ValidProcessor()
        asset_func = processor.get_assetable_func()
        result = asset_func(
            value=5, export=True, output_dir=self.output_dir, exist_ok=True
        )
        self.assertEqual(result, 10)

        # Test with invalid signature
        class InvalidProcessor:
            """A class with an invalid process signature."""

            def __init__(self):
                pass

            def process(self, value):
                """Process method missing required parameters."""
                return value * 2

        # This should raise InvalidProcessSignatureError
        invalid_processor = InvalidProcessor()
        with self.assertRaises(InvalidProcessSignatureError):
            validate_kwargs_signature(
                obj=invalid_processor,
                method_name="process",
                required_params=["export", "output_dir", "exist_ok"],
                exception_type=InvalidProcessSignatureError,
            )

    def test_missing_required_params(self):
        """Test that missing required parameters raises an error."""

        class PartialProcessor(ProcessorMixin):
            """A processor with a partially valid signature."""

            def __init__(self):
                ProcessorMixin.__init__(self)

            def process(self, value=None, export=False):
                """Process method missing some required parameters.

                Args:
                    value: The value to process
                    export: Whether to export the results

                Returns:
                    The processed value
                """
                return value * 2

        processor = PartialProcessor()
        with self.assertRaises(InvalidProcessSignatureError):
            processor.get_assetable_func()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Base class for exportable dataclass.
"""

import logging

from ...format.info import DataclassInfo
from ...format.base import Formatter
from .strategy.base import ExportStrategy

LOGGER = logging.getLogger(__name__)


class ExportableMixin:
    """Mixin class for exportable dataclass.

    This class provides the core functionality for exporting dataclass schemas
    to various formats using different export strategies.
    """

    _EXPORT_SETUP_ = None
    _EXPORTABLE_ = True

    @classmethod
    def to_schema_info(cls):
        """Convert the class definition to DataclassInfo."""
        return DataclassInfo.from_class(
            cls, hide_private=cls._EXPORT_SETUP_.hide_private
        )

    @classmethod
    def export(
        cls, output_dir, format_type="yaml", exist_ok=False, makedirs=False, **kwargs
    ):
        """Export the class schema definition in the specified format.

        Args:
            output_dir: Output directory
            format_type: Output format (yaml, json, etc.)
            exist_ok: Overwrite existing file
            makedirs: Create missing directories
            **kwargs: Additional formatting options
                - detailed: bool = True    # detailed mode
                - show_docs: bool = True   # Include docstrings
        """
        strategy = cls._get_export_strategy()
        strategy.export(cls, output_dir, format_type, exist_ok, makedirs, **kwargs)

    @classmethod
    def view_schema(cls, format_type="yaml", **kwargs):
        """Display the class schema definition.

        Args:
            format_type: Display format (yaml, json, etc.)
            **kwargs: Additional display options
        """
        schema_info = cls.to_schema_info()
        formatter = Formatter.get_registered(format_type)()
        formatter.update_settings(**kwargs)
        content = formatter.to_str(schema_info)
        formatter.display(content)

    @classmethod
    def _get_export_strategy(cls):
        """Get the appropriate export strategy for the decorated class."""
        strategy_name = cls._get_export_strategy_name()
        return ExportStrategy.get_registered(strategy_name)

    @classmethod
    def _get_export_strategy_name(cls):
        """Get the appropriate strategy for a class."""
        return cls._EXPORT_SETUP_.strategy



#!/usr/bin/env python3
# pylint: disable=too-few-public-methods
"""
Viewer exceptions.
"""


from ...exceptions import DataclassError


class ViewerError(DataclassError):
    """Base exception for viewer decorator."""


class ViewerSetupError(ViewerError):
    """Raised when there is an error in the Viewer setup configuration."""


class ViewerFormattingError(ViewerError):
    """Raised when there is an error during formatting operations."""



#!/usr/bin/env python3
"""
Sink mixin implementation.
"""

from ....utils.kwargs import validate_kwargs_signature
from ..component.sink import SinkComponent
from .exceptions import InvalidConsumeSignatureError, REQUIRE_KWARGS_SIGNATURE


class SinkMixin(SinkComponent):
    """Mixin that implements get_assetable_func for sinks."""

    def get_assetable_func(self):
        """Convert consume method to assetable function.

        Returns:
            A function that can be used as a dagster asset
        """

        validate_kwargs_signature(
            obj=self,
            method_name="consume",
            required_params=REQUIRE_KWARGS_SIGNATURE,
            exception_type=InvalidConsumeSignatureError,
        )

        def asset_func(**inputs):
            # Simply pass all inputs to consume
            return self.consume(**inputs)

        asset_func.__name__ = self.__class__.__name__
        return asset_func



#!/usr/bin/env python3
"""
Runner application configuration decorator.
"""

from pydantic.dataclasses import dataclass

from ....fallback.pydantic import is_pydantic_dataclass
from ....runner.workflow.config.base import WorkflowConfig
from ..wenv.validation import is_wenv
from .exceptions import RunnerAttributeError


def runner(_cls=None):
    """Decorator to ensure the configuration structure of a runner application.

    - Checks the presence of the workflow attribute: WorkflowConfig
    - Checks that workenv is a class decorated with @wenv
    - Adds utility methods for integration with Hydra
    """

    def wrap(cls):
        if not is_pydantic_dataclass(cls):
            cls = dataclass(cls)

        class RunnerAppConfig(cls):  # pylint: disable=C0115, R0903
            def __post_init__(self):
                if hasattr(super(), "__post_init__"):
                    super().__post_init__()

                if not hasattr(self, "workflow"):
                    raise RunnerAttributeError(
                        f"Missing required attribute 'workflow' in {self.__class__.__name__}. "
                        "Ensure you define a 'workflow' attribute in your runner app config dataclass."
                    )
                if not hasattr(self, "workenv"):
                    raise RunnerAttributeError(
                        f"Missing required attribute 'workenv' in {self.__class__.__name__}. "
                        "Ensure you define a 'workenv' attribute in your runner app config dataclass."
                    )

                if not isinstance(self.workflow, WorkflowConfig):
                    raise RunnerAttributeError(
                        f"Invalid type for 'workflow' in {self.__class__.__name__}. "
                        f"Expected 'WorkflowConfig', got {type(self.workflow).__name__}."
                    )
                if not is_wenv(self.workenv):
                    raise RunnerAttributeError(
                        f"Invalid work environment in {self.__class__.__name__}. "
                        "The 'workenv' attribute must be decorated with @wenv."
                    )

        RunnerAppConfig.__name__ = cls.__name__
        RunnerAppConfig.__qualname__ = cls.__qualname__
        RunnerAppConfig.__module__ = cls.__module__
        return RunnerAppConfig

    if _cls is None:
        return wrap
    return wrap(_cls)



#!/usr/bin/env python3
"""
Base class for custom workflow components.
"""

from abc import ABC, abstractmethod

from .....mixin.registrable import Registrable
from .....mixin.settings import SettingsMixin


class CustomComponent(ABC, Registrable, SettingsMixin):
    """Base class for all custom components.

    This is the base class that users should extend to create their own
    custom workflow components. It provides registration and settings functionality.
    """

    def __init__(self, settings, *, workenv):
        """Initialize the component.

        Args:
            settings: Component settings
        """
        Registrable.__init__(self)
        SettingsMixin.__init__(self, settings)
        self._workenv = workenv

    @classmethod
    @abstractmethod
    def init_from_config(cls, config, workenv=None):
        """Initialize the component from configuration.

        Args:
            config: Component configuration
            workenv: Optional work environment instance

        Returns:
            An instance of the component
        """



#!/usr/bin/env python3
"""
Setup for exportable.
"""

from dataclasses import dataclass


@dataclass
class ExportableSetup:
    """
    Setup configuration for exportable.

    Attributes:
        strategy (str): Export strategy
        hide_private (bool): Whether to hide private attributes during export
        stem_file (str): Base filename without extension for export.
            If None, the dataclass name will be used
    """

    strategy: str = "default"
    hide_private: bool = True
    stem_file: str = None



#!/usr/bin/env python3
"""
Dagster integration for workflow execution.
"""

import inspect
from functools import wraps

from ....optional.runner.dagster import (
    asset,
    AssetIn,
    Definitions,
    define_asset_job,
    mem_io_manager,
)
from ...workenv.component.exceptions import InvalidComponentError
from .base import BaseStep, BaseWorkflow


class DagsterStep(BaseStep):
    """Dagster-based workflow step."""

    def __init__(self, config, workenv):
        super().__init__(config, workenv)
        self._asset = None

    @property
    def asset(self):
        """Return dagster asset."""
        if self._asset is None:
            self._asset = self.get_asset()
        return self._asset

    def get_asset(self):
        """Get dagster asset."""
        asset_func = getattr(self.component, "get_assetable_func", None)
        if asset_func is None:
            raise InvalidComponentError(
                f"{self.component.__class__.__name__} is not a valid component for Dagster. "
                f"{self.component.__class__.__name__} as no attribute `get_assetable_func`."
            )

        def configured_func(**inputs):
            return asset_func()(
                export=self.save_results,
                output_dir=self.output_dir,
                exist_ok=self.exist_ok,
                **inputs,
            )

        return self._create_asset(configured_func)

    def _create_asset(self, func):
        """Create dagster asset."""
        deps = {}
        sig = inspect.signature(func)

        if sig.parameters and self.inputs:
            deps = {
                param: AssetIn(asset_name) for param, asset_name in self.inputs.items()
            }

        @asset(key=self.config.name, ins=deps)
        @wraps(func)
        def wrapped(**inputs):
            return func(**inputs)

        return wrapped

    def execute(self, **inputs):
        """Execute the step using its dagster asset."""
        return self.asset(**inputs)


class DagsterWorkflow(BaseWorkflow):
    """Dagster-based workflow implementation."""

    def _initialize_steps(self, workenv):
        """Initialize dagster workflow steps."""
        self._steps = [DagsterStep(step_cfg, workenv) for step_cfg in self.config.steps]

    @property
    def assets(self):
        """Return workflow assets."""
        return [step.asset for step in self.steps]

    def execute(self):
        """Execute the dagster workflow."""
        defs = Definitions(
            assets=self.assets,
            jobs=[define_asset_job(name=self.config.name)],
            resources={"io_manager": mem_io_manager},
        )

        return defs.get_job_def(self.config.name).execute_in_process(
            resources={"io_manager": mem_io_manager}
        )



#!/usr/bin/env python3
"""Convert Utility functions."""

from collections import OrderedDict
from collections.abc import Iterable


def to_dict_recursive(obj):
    """Convert any object containing OrderedDict to standard dict recursively.

    This function handles nested data structures and converts OrderedDict
    to standard dict while preserving the order of keys. It can process:
    - OrderedDict
    - dict
    - list
    - tuple
    - any other type (returned as is)

    Args:
        obj: The object to convert, can be of any type

    Returns:
        The converted object with all OrderedDict replaced by standard dict

    Examples:
        >>> data = OrderedDict([('a', 1), ('b', OrderedDict([('c', 2)]))])
        >>> result = to_dict_recursive(data)
        >>> isinstance(result, dict)
        True
        >>> isinstance(result['b'], dict)
        True
    """
    if isinstance(obj, (OrderedDict, dict)):
        return {key: to_dict_recursive(value) for key, value in obj.items()}
    if isinstance(obj, list):
        return [to_dict_recursive(item) for item in obj]
    if isinstance(obj, tuple):
        return tuple(to_dict_recursive(item) for item in obj)
    return obj


def ensure_list(obj):
    """
    Ensures that the input is returned as a list.

    - If the input is already a list, it is returned unchanged.
    - If the input is an iterable (e.g., tuple, set, numpy array), it is converted to a list.
    - If the input is a scalar or a string/bytes, it is wrapped in a single-element list.

    Args:
        obj (Any): The object to convert to a list.

    Returns:
        List: A list version of the input object.
    """
    if isinstance(obj, list):
        return obj
    if isinstance(obj, Iterable) and not isinstance(obj, (str, bytes)):
        return list(obj)
    return [obj]



#!/usr/bin/env python3
"""
Default export strategy for simple configurations.
"""

import pathlib
import logging

from ..base import ExportStrategy
from .....format.base import Formatter
from ......utils.export import export_string, create_directory


LOGGER = logging.getLogger(__name__)


class DefaultExportStrategy(ExportStrategy, register_name="default"):
    """Default export strategy for dataclasses."""

    @classmethod
    def export(  # pylint: disable=too-many-arguments
        cls,
        data_cls,
        output_dir,
        format_type="yaml",
        exist_ok=False,
        makedirs=False,
        **kwargs,
    ):
        """Export the configuration."""
        formatter = Formatter.get_registered(format_type)()
        formatter.update_settings(**kwargs)
        schema_info = data_cls.to_schema_info()
        content = formatter.to_str(schema_info)

        base_name = (
            # pylint: disable=protected-access
            data_cls._EXPORT_SETUP_.stem_file
        )
        if base_name is None:
            base_name = data_cls.__name__.lower()

        filename = f"{base_name}.{format_type}"
        if makedirs:
            output_dir = create_directory(output_dir)
        else:
            output_dir = pathlib.Path(output_dir).resolve()
        output_path = output_dir / filename
        export_string(content, output_path, exist_ok=exist_ok)
        LOGGER.info("Saved schema to %s", output_path)



#!/usr/bin/env python3
"""
Registry decorator for dataclass.
"""

import logging

from ..base import BaseDecorator

from .exceptions import RegistrySetupError
from .setup import RegistrySetup
from .handlers import TypeRegistry, SettingsHandler

LOGGER = logging.getLogger(__name__)


def registry(base_cls, _cls=None, *, setup=None, **kwargs):
    """Exportable decorator for registry-based schema definitions.

    Args:
        _cls: The class to decorate
        base_cls: The base class for the registry
        setup: Optional. Registry setup configuration.
        kwargs: Optional setup overrides

    Example:
        @registry(base_cls=MyBaseClass, setup=MySetup)
        class MyClassConfig:
            name: str
            settings: Dict[str, Any]
    """

    def wrap(cls):
        try:
            cls, setup_instance = BaseDecorator.decorate_class(
                cls, RegistrySetup, setup, **kwargs
            )
        except ValueError as err:
            raise RegistrySetupError(str(err)) from err

        class RegistryClass(cls):  # pylint: disable=R0903, C0115
            cls._SETUP_ = setup_instance
            cls._REG_BASE_CLASS_ = base_cls
            cls._REGISTRY_ = True

            def _handle_parent_post_init(self):
                """Handle parent class post-init if it exists."""
                post_init = getattr(super(), "__post_init__", None)
                if callable(post_init):
                    post_init()  # pylint: disable=not-callable

            def _configure_type_registration(self):
                """Configure type registration and get registered class."""
                type_value = TypeRegistry.validate_type(self, self._SETUP_)
                return TypeRegistry.get_registration(self._REG_BASE_CLASS_, type_value)

            def _setup_settings(self, reg_cls, type_value):
                """Setup settings configuration."""
                SettingsHandler.configure_settings(
                    self, reg_cls, type_value, self._SETUP_
                )

            def __post_init__(self):
                self._handle_parent_post_init()
                reg_cls = self._configure_type_registration()
                type_value = getattr(self, self._SETUP_.register_name_attr)
                self._setup_settings(reg_cls, type_value)

            @classmethod
            def reload_registry(cls, clear=False):
                """Reload registry.

                Args:
                    clear: If True, clears all current registrations before reloading
                """
                LOGGER.debug("Reloading registry from %s", cls.__name__)
                # pylint: disable=protected-access
                submod = cls._REG_BASE_CLASS_._TYPE_SUBMODULE
                cls._REG_BASE_CLASS_.reload(submod=submod, clear=clear)

        BaseDecorator.preserve_cls_metadata(RegistryClass, cls)
        return RegistryClass

    if _cls is None:
        return wrap
    return wrap(_cls)



#!/usr/bin/env python3
"""
Exportable exceptions.
"""


from ...exceptions import DataclassError


class ExportableError(DataclassError):
    """Base exception for exportable decorator."""


class ExportableSetupError(ExportableError):
    """Raised when there is an error in the Exportable setup configuration."""



#!/usr/bin/env python3
"""
Logging utilities.
"""

import logging
from contextlib import contextmanager


@contextmanager
def temporary_console_logger(logger, level=logging.DEBUG):
    """
    Temporarily adds a console handler to the logger and adjusts its level.
    Args:
        logger: Logger instance.
        level: Temporary log level (e.g., logging.DEBUG, logging.INFO).
    """
    # Save previous values
    previous_level = logger.level

    # Configure the logger temporarily
    logger.setLevel(level)
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    formatter = logging.Formatter("%(levelname)s: %(message)s")
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    try:
        yield logger
    finally:
        # Restore previous state
        logger.removeHandler(console_handler)
        logger.setLevel(previous_level)



#!/usr/bin/env python3
"""Unit tests for workflow source components."""

import tempfile
import unittest
from pathlib import Path
from pydantic.dataclasses import dataclass

from pypassist.dataclass.decorators.viewer import viewer
from pypassist.runner.workenv.mixin.source import SourceMixin
from pypassist.mixin.settings import SettingsMixin
from pypassist.utils.export import export_string
from pypassist.runner.workenv.mixin.exceptions import InvalidProvideSignatureError
from pypassist.utils.kwargs import validate_kwargs_signature


class TestSource(unittest.TestCase):
    """Test cases for workflow source components."""

    def setUp(self):
        """Set up test fixtures."""
        self.temp_dir = tempfile.mkdtemp()
        self.output_dir = Path(self.temp_dir)

        # Create a test file for file source
        self.test_file = Path(self.temp_dir) / "test_input.txt"
        with open(self.test_file, "w", encoding="utf-8") as f:
            f.write("line1\nline2\nline3\n")

    def test_basic_source(self):
        """Test basic source functionality."""

        @viewer
        @dataclass
        class FileSourceSettings:
            """Settings for file source."""

            filename: str = "input.txt"
            strip_lines: bool = True

        class FileSource(SettingsMixin, SourceMixin):
            """A simple source that reads data from a file."""

            SETTINGS_DATACLASS = FileSourceSettings

            def __init__(self, settings=None):
                if settings is None:
                    settings = FileSourceSettings()
                SettingsMixin.__init__(self, settings)
                SourceMixin.__init__(self)

            def provide(
                self,
                filepath=None,
                export=False,
                output_dir=None,
                exist_ok=True,
                **kwargs,
            ):
                """Provide data from a file.

                Args:
                    filepath: Path to the file to read (overrides settings.filename)
                    export: Whether to export the results
                    output_dir: Directory to export results to
                    exist_ok: Whether to overwrite existing files
                    **kwargs: Additional keyword arguments

                Returns:
                    The file contents as a list of lines
                """
                if filepath is None:
                    raise ValueError("Filepath parameter is required")

                # Read the file
                with open(filepath, "r", encoding="utf-8") as f:
                    if self.settings.strip_lines:
                        lines = [line.strip() for line in f.readlines()]
                    else:
                        lines = f.readlines()

                if export and output_dir:
                    # Export the data
                    result_content = "\n".join([line.strip() for line in lines])
                    result_filepath = (
                        output_dir / f"{self.__class__.__name__.lower()}_result.txt"
                    )
                    export_string(
                        result_content,
                        filepath=result_filepath,
                        exist_ok=exist_ok,
                        makedirs=True,
                    )

                    # Export settings
                    settings_content = self.settings.to_str(format_type="yaml")
                    settings_filepath = (
                        output_dir / f"{self.__class__.__name__.lower()}_settings.txt"
                    )
                    export_string(
                        settings_content,
                        filepath=settings_filepath,
                        exist_ok=exist_ok,
                        makedirs=True,
                    )

                return lines

        # Test basic source
        source = FileSource()
        lines = source.provide(filepath=self.test_file)
        self.assertEqual(lines, ["line1", "line2", "line3"])

        # Test with custom settings
        settings = FileSourceSettings(strip_lines=False)
        source = FileSource(settings)
        lines = source.provide(filepath=self.test_file)
        self.assertEqual(lines, ["line1\n", "line2\n", "line3\n"])

        # Test asset generation
        asset_func = source.get_assetable_func()
        result = asset_func(
            filepath=self.test_file, export=True, output_dir=self.output_dir
        )
        self.assertEqual(result, ["line1\n", "line2\n", "line3\n"])

        # Verify exported files
        result_file = self.output_dir / "filesource_result.txt"
        settings_file = self.output_dir / "filesource_settings.txt"
        self.assertTrue(result_file.exists())
        self.assertTrue(settings_file.exists())
        self.assertEqual(result_file.read_text(), "line1\nline2\nline3")

    def test_provide_signature_validation(self):
        """Test provide signature validation."""

        # Test with valid signature
        class ValidSource(SourceMixin):
            """A source with a valid provide signature."""

            def __init__(self):
                SourceMixin.__init__(self)

            def provide(self, data=None, export=False, output_dir=None, exist_ok=True):
                """Provide with valid signature including all required parameters.

                Args:
                    data: The data to provide
                    export: Whether to export the results
                    output_dir: Directory to export results to
                    exist_ok: Whether to overwrite existing files

                Returns:
                    The provided data
                """
                return data

        # This should work fine
        source = ValidSource()
        asset_func = source.get_assetable_func()
        result = asset_func(
            data="Test data", export=True, output_dir=self.output_dir, exist_ok=True
        )
        self.assertEqual(result, "Test data")

        # Test with invalid signature
        class InvalidSource:
            """A class with an invalid provide signature."""

            def __init__(self):
                pass

            def provide(self, data):
                """Provide method missing required parameters."""
                return data

        # This should raise InvalidProvideSignatureError
        invalid_source = InvalidSource()
        with self.assertRaises(InvalidProvideSignatureError):
            validate_kwargs_signature(
                obj=invalid_source,
                method_name="provide",
                required_params=["export", "output_dir", "exist_ok"],
                exception_type=InvalidProvideSignatureError,
            )

    def test_missing_required_params(self):
        """Test that missing required parameters raises an error."""

        class PartialSource(SourceMixin):
            """A source with a partially valid signature."""

            def __init__(self):
                SourceMixin.__init__(self)

            def provide(self, data=None, export=False):
                """Provide method missing some required parameters.

                Args:
                    data: The data to provide
                    export: Whether to export the results

                Returns:
                    The provided data
                """
                return data

        source = PartialSource()
        with self.assertRaises(InvalidProvideSignatureError):
            source.get_assetable_func()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Exportable decorator for dataclass.
"""

import logging

from ..base import BaseDecorator
from .setup import ExportableSetup
from .exportable import ExportableMixin
from .exceptions import ExportableSetupError

LOGGER = logging.getLogger(__name__)


def exportable(_cls=None, *, setup=None, **kwargs):
    """Exportable decorator for enhanced data export capabilities.

    Args:
        _cls: The class to decorate
        setup: ExportableSetup instance containing export configuration
        kwargs: Setup fields. Will override corresponding field in setup.

    Example:
        @exportable
        class MyConfig:
            name: str
            value: int
    """

    def wrap(cls):
        try:
            cls, setup_instance = BaseDecorator.decorate_class(
                cls, ExportableSetup, setup, **kwargs
            )
        except ValueError as err:
            raise ExportableSetupError(str(err)) from err

        class ExportableClass(cls, ExportableMixin):  # pylint: disable=R0903, C0115
            cls._EXPORT_SETUP_ = setup_instance

        BaseDecorator.preserve_cls_metadata(ExportableClass, cls)
        return ExportableClass

    if _cls is None:
        return wrap
    return wrap(_cls)



#!/usr/bin/env python3
"""
Base configuration types for workflow module.
"""

import pathlib
from typing import Optional, Any

from pydantic.dataclasses import dataclass

from ....dataclass.decorators.wenv.validation import is_wenv
from ....dataclass.decorators.exportable.decorator import exportable
from ....fallback.typing import Dict, List
from ...workenv.component.exceptions import InvalidComponentError


@dataclass
class StepConfig:
    """Step configuration."""

    component_path: str  # Format: "category.name" (e.g. "sources.file")
    name: Optional[str] = None
    inputs: Optional[Dict[str, Any]] = None
    description: Optional[str] = None
    save_results: bool = True
    exist_ok: bool = False
    _output_dir: Optional[str] = None
    _workflow_output_dir: Optional[str] = None

    def __post_init__(self):
        if self.name is None:
            self.name = self.component_path.split(".")[-1]

    @property
    def output_dir(self):
        """Return Path to output directory."""
        if self._output_dir:
            return pathlib.Path(self._output_dir).resolve()
        return pathlib.Path(self._workflow_output_dir).resolve() / self.name

    def get_component(self, workenv):
        """Get component from nested sections.

        Args:
            workenv: Work environment instance

        Returns:
            Component instance
        """
        component_path = self.component_path
        *attr_path, dict_key = component_path.split(".")

        section_name = attr_path[0]
        current = getattr(workenv.config, section_name)

        if is_wenv(current):
            current = getattr(workenv, f"{section_name}")
            remaining_path = ".".join([*attr_path[1:], dict_key])
            return current.get_component(remaining_path)

        if not isinstance(current, dict):
            ## Direct component instance
            return getattr(workenv, f"{section_name}")

        component_dict = getattr(workenv, f"{section_name}")
        if dict_key not in component_dict:
            raise InvalidComponentError(
                f"{self.__class__.__name__}: Component '{dict_key}' not found in section '{section_name}'. "
                f"Available components: {', '.join(component_dict.keys())}"
            )
        return component_dict[dict_key]


@exportable(stem_file="template")
@dataclass
class WorkflowConfig:
    """Workflow configuration."""

    name: str
    steps: List[Dict]
    output_dir: str
    save_results: bool = True
    exist_ok: bool = False

    def __post_init__(self):
        step_defaults = {"exist_ok": self.exist_ok, "save_results": self.save_results}
        self.steps = [
            StepConfig(
                **{**step_defaults, **step, "_workflow_output_dir": self.output_dir}
            )
            for step in self.steps
        ]



"""
Custom work environment configuration.
"""

from pydantic.dataclasses import dataclass, Field

from .....fallback.typing import Dict
from .....dataclass.decorators.exportable.decorator import exportable
from .....dataclass.decorators.wenv.decorator import wenv
from .processor import CustomProcessorConfig
from .sink import CustomSinkConfig
from .source import CustomSourceConfig


@wenv
@exportable(strategy="wenv", stem_file="workenv_defaults")
@dataclass
class CustomWorkEnvConfig:
    """
    Custom work environment configuration.
    """

    processors: Dict[str, CustomProcessorConfig] = Field(default_factory=dict)
    sink: Dict[str, CustomSinkConfig] = Field(default_factory=dict)
    sources: Dict[str, CustomSourceConfig] = Field(default_factory=dict)



#!/usr/bin/env python3
"""Progress bar."""

import logging

from ..utils.ipynb import is_in_ipynb

LOGGER = logging.getLogger(__name__)


# Fallback class if tqdm is not available
class DummyTqdm:
    """A no-op progress bar fallback when tqdm is not available."""

    def __init__(self, *args, **kwargs):
        self.total = kwargs.get("total", None)
        self.n = 0

    def update(self, n=1):
        """
        Mimic tdqm update method.
        """
        self.n += n

    def close(self):
        """
        Mimic tdqm close method.
        """

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.close()


# Try to import tqdm, use dummy if not available
try:
    # pylint: disable=W0611
    if is_in_ipynb():
        from tqdm.notebook import tqdm  # For Jupyter Notebooks

        LOGGER.debug("Using tqdm for notebooks.")
    else:
        from tqdm import tqdm  # For normal environments

        LOGGER.debug("Using standard tqdm.")
except ImportError:
    tqdm = DummyTqdm  # pylint: disable=C0103
    LOGGER.debug("tqdm not available, progress bar disabled.")



#!/usr/bin/env python
"""
Utilities for working with Jupyter notebooks.
"""


def is_in_ipynb():
    """Check if the code is running in a Jupyter Notebook."""
    try:
        from IPython import get_ipython  # pylint: disable=import-outside-toplevel

        return get_ipython() is not None and "IPKernelApp" in get_ipython().config
    except (ImportError, AttributeError):
        return False



#!/usr/bin/env python3
"""
JSON formatter.
"""

import logging

import json
from pydantic.dataclasses import dataclass

from ..base import Formatter
from ....utils.ipynb import is_in_ipynb

LOGGER = logging.getLogger(__name__)


@dataclass
class JsonFormatterSettings:
    """JSON formatter settings."""

    indent: int = 4
    sort_keys: bool = False
    detailed: bool = False  # detailed mode
    show_docs: bool = False  # whether to include doc in detailed mode


class JsonFormatter(Formatter, register_name="json"):
    """JSON formatter."""

    SETTINGS_DATACLASS = JsonFormatterSettings

    def __init__(self, settings=None):
        if settings is None:
            settings = JsonFormatterSettings()
        super().__init__(settings)

    def to_str(self, class_info):
        """Format DataclassInfo into JSON string."""
        if self.settings.detailed:
            data = self._build_detailed(class_info)
        else:
            data = self._build_simple(class_info)

        return json.dumps(
            data,
            indent=self.settings.indent,
            sort_keys=self.settings.sort_keys,
        )

    def display(self, content):
        """Display the formatted JSON content."""
        if is_in_ipynb():
            # pylint: disable=import-outside-toplevel
            from IPython.display import (
                display_json,
            )

            LOGGER.debug("Displaying JSON content in Jupyter Notebook.")
            display_json(content, raw=True)
        else:
            LOGGER.debug("Displaying JSON content in terminal.")
            print(content)

    def to_serialized(self, class_info):
        """Convert DataclassInfo to a dictionary for JSON serialization."""
        if self.settings.detailed:
            return self._build_detailed(class_info)
        return self._build_simple(class_info)

    def _build_simple(self, class_info):
        """Build a simple name/value representation for direct instantiation.
        This is the default format, suitable for serialization/deserialization.
        """
        result = {}
        for field_info in class_info.fields:
            if isinstance(field_info.value, type(class_info)):
                result[field_info.name] = self._build_simple(field_info.value)
            else:
                result[field_info.name] = field_info.value
        return result

    def _build_detailed(self, class_info):
        """Build a detailed schema representation including metadata."""
        result = {"name": class_info.name, "fields": {}}

        if self.settings.show_docs:
            result["description"] = class_info.description

        for field_info in class_info.fields:
            field_data = {
                "type": field_info.type_info,
            }

            if self.settings.show_docs:
                field_data["description"] = field_info.description

            if isinstance(field_info.value, type(class_info)):
                field_data["value"] = self._build_detailed(field_info.value)
            else:
                field_data["value"] = field_info.value

            result["fields"][field_info.name] = field_data

        return result



#!/usr/bin/env python3
"""
Source component protocol for workflow module.
"""

from .base import WorkflowComponent
from ....fallback.protocol import runtime_checkable, Protocol


@runtime_checkable
class SourceComponent(WorkflowComponent, Protocol):
    """Protocol for components that provide data."""

    def provide(self, **kwargs):
        """Provide input data with flexible signature.

        All implementations should support these keyword arguments:
            export (bool): Whether to export the results
            output_dir (str): Directory to export results to
            exist_ok (bool): Whether to overwrite existing files

        Args:
            **kwargs: Arbitrary keyword arguments

        Returns:
            Any: The provided data
        """



#!/usr/bin/env python3
"""Exportable decorator package."""

from .decorator import exportable
from .setup import ExportableSetup
from .exceptions import ExportableSetupError

__all__ = ["exportable", "ExportableSetup", "ExportableSetupError"]
